import abc
import contextlib
import inspect
import threading
import typing
from contextlib import asynccontextmanager, contextmanager
from operator import attrgetter

import typing_extensions
from typing_extensions import override

from that_depends.entities.resource_context import ResourceContext
from that_depends.providers.mixin import SupportsTeardown


T_co = typing.TypeVar("T_co", covariant=True)
R = typing.TypeVar("R")
P = typing.ParamSpec("P")
ResourceCreatorType: typing.TypeAlias = typing.Callable[
    P,
    typing.Iterator[T_co] | typing.AsyncIterator[T_co] | typing.ContextManager[T_co] | typing.AsyncContextManager[T_co],
]


class AbstractProvider(typing.Generic[T_co], abc.ABC):
    """Base class for all providers."""

    def __init__(self) -> None:
        """Create a new provider."""
        super().__init__()
        self._children: set[AbstractProvider[typing.Any]] = set()
        self._parents: set[AbstractProvider[typing.Any]] = set()
        self._override: typing.Any = None
        self._lock = threading.Lock()

    def _register(self, candidates: typing.Iterable[typing.Any]) -> None:
        """Register current provider as child.

        Args:
            candidates: iterable of potential parent providers.

        Returns:
            None

        """
        for candidate in candidates:
            if isinstance(candidate, AbstractProvider):
                candidate.add_child_provider(self)
                self._parents.add(candidate)

    def _deregister(self, candidates: typing.Iterable[typing.Any]) -> None:
        """Deregister current provider as child.

        Args:
            candidates: iterable of potential parent providers.

        Returns:
            None

        """
        for candidate in candidates:
            if isinstance(candidate, AbstractProvider) and self in candidate._children:  # noqa: SLF001
                candidate.remove_child_provider(self)
                self._parents.discard(candidate)

    def add_child_provider(self, provider: "AbstractProvider[typing.Any]") -> None:
        """Add a child provider to the current provider.

        Args:
            provider: provider to add as a child.

        Returns:
            None

        """
        with self._lock:
            self._children.add(provider)

    def remove_child_provider(self, provider: "AbstractProvider[typing.Any]") -> None:
        """Remove a child provider from the current provider.

        Args:
            provider: provider to remove as a child.

        Returns:
            None

        """
        with self._lock:
            self._children.discard(provider)

    def __deepcopy__(self, *_: object, **__: object) -> typing_extensions.Self:
        """Hack for Litestar to prevent cloning object.

        More info here https://github.com/modern-python/that-depends/issues/119.
        """
        return self

    def __getattr__(self, attr_name: str) -> typing.Any:  # noqa: ANN401
        """Get an attribute from the resolve object.

        Args:
            attr_name: name of attribute to get.

        Returns:
            An `AttrGetter` provider that will get the attribute after resolving the current provider.

        """
        if attr_name.startswith("_"):
            msg = f"'{type(self)}' object has no attribute '{attr_name}'"
            raise AttributeError(msg)
        return AttrGetter(provider=self, attr_name=attr_name)

    @abc.abstractmethod
    async def resolve(self) -> T_co:
        """Resolve dependency asynchronously."""

    @abc.abstractmethod
    def resolve_sync(self) -> T_co:
        """Resolve dependency synchronously."""

    async def __call__(self) -> T_co:
        """Resolve dependency asynchronously."""
        return await self.resolve()

    def override_sync(
        self, mock: object, tear_down_children: bool = False, propagate: bool = True, raise_on_async: bool = False
    ) -> None:
        """Override the provider with a mock object.

        Args:
            mock: object to resolve while the provider is overridden.
            tear_down_children: tear down child providers.
            raise_on_async: raise if tear down requires async context.
            propagate: propagate teardown.

        Returns:
            None

        """
        self._override = mock
        if tear_down_children:
            eligible_children = [child for child in self._children if isinstance(child, SupportsTeardown)]
            for child in eligible_children:
                child.tear_down_sync(propagate=propagate, raise_on_async=raise_on_async)

    async def override(self, mock: object, tear_down_children: bool = False, propagate: bool = True) -> None:
        """Override the provider with a mock object.

        Args:
            mock: object to resolve while the provider is overridden.
            tear_down_children: tear down child providers.
            propagate: propagate teardown.

        Returns:
            None

        """
        self._override = mock
        if tear_down_children:
            eligible_children = [child for child in self._children if isinstance(child, SupportsTeardown)]
            for child in eligible_children:
                await child.tear_down(propagate=propagate)

    @contextmanager
    def override_context_sync(self, mock: object) -> typing.Iterator[None]:
        """Override the provider with a mock object temporarily.

        Args:
            mock: object to resolve while the provider is overridden.

        Returns:
            None

        """
        self.override_sync(mock)
        try:
            yield
        finally:
            self.reset_override_sync()

    @asynccontextmanager
    async def override_context(self, mock: object) -> typing.AsyncIterator[None]:
        """Override the provider with a mock object temporarily.

        Args:
            mock: object to resolve while the provider is overridden.

        Returns:
            None

        """
        await self.override(mock)
        try:
            yield
        finally:
            self.reset_override_sync()

    def reset_override_sync(
        self, tear_down_children: bool = False, propagate: bool = True, raise_on_async: bool = False
    ) -> None:
        """Reset the provider to its original state.

        Use this is you have previously called `override` or `override_context`
        to reset the provider to its original state.

        Args:
            tear_down_children: tear down all child providers.
            raise_on_async: raise if an async teardown is necessary.
            propagate: propagate tear downs.

        Returns:
            None

        """
        self._override = None
        if tear_down_children:
            eligible_children = [child for child in self._children if isinstance(child, SupportsTeardown)]
            for child in eligible_children:
                child.tear_down_sync(propagate=propagate, raise_on_async=raise_on_async)

    async def reset_override(self, tear_down_children: bool = False, propagate: bool = True) -> None:
        """Reset the provider to its original state.

        Args:
            tear_down_children: tear down all child providers.
            propagate: propagate tear downs.

        Returns:
            None

        """
        self._override = None
        if tear_down_children:
            eligible_children = [child for child in self._children if isinstance(child, SupportsTeardown)]
            for child in eligible_children:
                await child.tear_down(propagate=propagate)

    @property
    def cast(self) -> T_co:
        """Returns self, but cast to the type of the provided value.

        This helps to pass providers as input to other providers while avoiding type checking errors:

        Example:
            class A: ...

            def create_b(a: A) -> B: ...

            class Container(BaseContainer):
                a_factory = Factory(A)
                b_factory1 = Factory(create_b, a_factory)  # works, but mypy (or pyright, etc.) will complain
                b_factory2 = Factory(create_b, a_factory.cast)  # works and passes type checking

        """
        return typing.cast(T_co, self)

    async def _tear_down_children(self) -> None:
        """Tear down all child providers."""
        eligible_children = [child for child in self._children if isinstance(child, SupportsTeardown)]
        for child in eligible_children:
            await child.tear_down()

    def _tear_down_children_sync(self, propagate: bool = True, raise_on_async: bool = True) -> None:
        """Tear down all child providers."""
        eligible_children = [child for child in self._children if isinstance(child, SupportsTeardown)]
        for child in eligible_children:
            child.tear_down_sync(raise_on_async=raise_on_async, propagate=propagate)


class AbstractResource(AbstractProvider[T_co], abc.ABC):
    """Base class for Resource providers."""

    def __init__(
        self,
        creator: ResourceCreatorType[P, T_co],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        """Create a new resource.

        Args:
            creator: sync or async iterator or context manager that yields resource.
            *args: positional arguments to pass to the creator.
            **kwargs: keyword arguments to pass to the creator.


        """
        super().__init__()
        self._creator: typing.Any

        if inspect.isasyncgenfunction(creator):
            self._is_async = True
            self._creator = contextlib.asynccontextmanager(creator)
        elif inspect.isgeneratorfunction(creator):
            self._is_async = False
            self._creator = contextlib.contextmanager(creator)
        elif isinstance(creator, type) and issubclass(creator, typing.AsyncContextManager):
            self._is_async = True
            self._creator = creator
        elif isinstance(creator, type) and issubclass(creator, typing.ContextManager):
            self._is_async = False
            self._creator = creator
        else:
            msg = "Unsupported resource type"
            raise TypeError(msg)
        self._args = args
        self._kwargs = kwargs

    def _register_arguments(self) -> None:
        self._register(self._args)
        self._register(self._kwargs.values())

    def _deregister_arguments(self) -> None:
        self._deregister(self._args)
        self._deregister(self._kwargs.values())

    @abc.abstractmethod
    def _fetch_context(self) -> ResourceContext[T_co]: ...

    @override
    async def resolve(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        context = self._fetch_context()

        # lock to prevent race condition while resolving
        async with context.asyncio_lock:
            if context.instance is not None:
                return context.instance

            self._register_arguments()

            cm: typing.ContextManager[T_co] | typing.AsyncContextManager[T_co] = self._creator(
                *[await x.resolve() if isinstance(x, AbstractProvider) else x for x in self._args],
                **{k: await v.resolve() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()},
            )

            if isinstance(cm, typing.AsyncContextManager):
                context.context_stack = contextlib.AsyncExitStack()
                context.instance = await context.context_stack.enter_async_context(cm)

            elif isinstance(cm, typing.ContextManager):
                context.context_stack = contextlib.ExitStack()
                context.instance = context.context_stack.enter_context(cm)

            else:  # pragma: no cover
                typing.assert_never(cm)

        return context.instance

    @override
    def resolve_sync(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        context = self._fetch_context()

        # lock to prevent race condition while resolving
        with context.threading_lock:
            if context.instance is not None:
                return context.instance

            if self._is_async:
                msg = "AsyncResource cannot be resolved synchronously"
                raise RuntimeError(msg)

            self._register_arguments()

            cm = self._creator(
                *[x.resolve_sync() if isinstance(x, AbstractProvider) else x for x in self._args],
                **{k: v.resolve_sync() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()},
            )
            context.context_stack = contextlib.ExitStack()
            context.instance = context.context_stack.enter_context(cm)

            return context.instance


def _get_value_from_object_by_dotted_path(obj: typing.Any, path: str) -> typing.Any:  # noqa: ANN401
    attribute_getter = attrgetter(path)
    return attribute_getter(obj)


class AttrGetter(
    AbstractProvider[T_co],
):
    """Provides an attribute after resolving the wrapped provider."""

    __slots__ = "_attrs", "_provider"

    def __init__(self, provider: AbstractProvider[T_co], attr_name: str) -> None:
        """Create a new AttrGetter instance.

        Args:
            provider: provider to wrap.
            attr_name: attribute name to resolve when the provider is resolved.

        """
        super().__init__()
        self._provider = provider
        self._attrs = [attr_name]

    @override
    def __getattr__(self, attr: str) -> "AttrGetter[T_co]":
        if attr.startswith("_"):
            msg = f"'{type(self)}' object has no attribute '{attr}'"
            raise AttributeError(msg)
        self._attrs.append(attr)
        return self

    @override
    async def resolve(self) -> typing.Any:
        resolved_provider_object = await self._provider.resolve()
        attribute_path = ".".join(self._attrs)
        return _get_value_from_object_by_dotted_path(resolved_provider_object, attribute_path)

    @override
    def resolve_sync(self) -> typing.Any:
        resolved_provider_object = self._provider.resolve_sync()
        attribute_path = ".".join(self._attrs)
        return _get_value_from_object_by_dotted_path(resolved_provider_object, attribute_path)

    @override
    def add_child_provider(self, provider: "AbstractProvider[typing.Any]") -> None:
        self._provider.add_child_provider(provider)



import datetime
import typing

import pytest

from tests import container
from that_depends import BaseContainer, providers
from that_depends.providers.mixin import CannotTearDownSyncError


async def test_batch_providers_overriding() -> None:
    async_resource_mock = datetime.datetime.fromisoformat("2023-01-01")
    sync_resource_mock = datetime.datetime.fromisoformat("2024-01-01")
    async_factory_mock = datetime.datetime.fromisoformat("2025-01-01")
    simple_factory_mock = container.SimpleFactory(dep1="override", dep2=999)
    singleton_mock = container.SingletonFactory(dep1=False)
    object_mock = object()

    providers_for_overriding = {
        "async_resource": async_resource_mock,
        "sync_resource": sync_resource_mock,
        "simple_factory": simple_factory_mock,
        "singleton": singleton_mock,
        "async_factory": async_factory_mock,
        "object": object_mock,
    }

    with container.DIContainer.override_providers_sync(providers_for_overriding):
        await container.DIContainer.simple_factory()
        dependent_factory = await container.DIContainer.dependent_factory()
        singleton = await container.DIContainer.singleton()
        async_factory = await container.DIContainer.async_factory()
        obj = await container.DIContainer.object()

    assert dependent_factory.simple_factory.dep1 == simple_factory_mock.dep1
    assert dependent_factory.simple_factory.dep2 == simple_factory_mock.dep2
    assert dependent_factory.sync_resource == sync_resource_mock
    assert dependent_factory.async_resource == async_resource_mock
    assert singleton is singleton_mock
    assert async_factory is async_factory_mock
    assert obj is object_mock

    assert (await container.DIContainer.async_resource()) != async_resource_mock


async def test_batch_providers_overriding_sync_resolve() -> None:
    async_resource_mock = datetime.datetime.fromisoformat("2023-01-01")
    sync_resource_mock = datetime.datetime.fromisoformat("2024-01-01")
    simple_factory_mock = container.SimpleFactory(dep1="override", dep2=999)
    singleton_mock = container.SingletonFactory(dep1=False)
    object_mock = object()

    providers_for_overriding = {
        "async_resource": async_resource_mock,
        "sync_resource": sync_resource_mock,
        "simple_factory": simple_factory_mock,
        "singleton": singleton_mock,
        "object": object_mock,
    }

    with container.DIContainer.override_providers_sync(providers_for_overriding):
        container.DIContainer.simple_factory.resolve_sync()
        await container.DIContainer.async_resource.resolve()
        dependent_factory = container.DIContainer.dependent_factory.resolve_sync()
        singleton = container.DIContainer.singleton.resolve_sync()
        obj = container.DIContainer.object.resolve_sync()

    assert dependent_factory.simple_factory.dep1 == simple_factory_mock.dep1
    assert dependent_factory.simple_factory.dep2 == simple_factory_mock.dep2
    assert dependent_factory.sync_resource == sync_resource_mock
    assert dependent_factory.async_resource == async_resource_mock
    assert singleton is singleton_mock
    assert obj is object_mock

    assert container.DIContainer.sync_resource.resolve_sync() != sync_resource_mock


def test_providers_overriding_with_context_manager() -> None:
    simple_factory_mock = container.SimpleFactory(dep1="override", dep2=999)

    with container.DIContainer.simple_factory.override_context_sync(simple_factory_mock):
        assert container.DIContainer.simple_factory.resolve_sync() is simple_factory_mock

    assert container.DIContainer.simple_factory.resolve_sync() is not simple_factory_mock


async def test_providers_overriding_with_async_context_manager() -> None:
    simple_factory_mock = container.SimpleFactory(dep1="override", dep2=999)

    async with container.DIContainer.simple_factory.override_context(simple_factory_mock):
        assert container.DIContainer.simple_factory.resolve_sync() is simple_factory_mock

    assert container.DIContainer.simple_factory.resolve_sync() is not simple_factory_mock


def test_providers_overriding_fail_with_unknown_provider() -> None:
    unknown_provider_name = "unknown_provider_name"
    match = f"Provider with name {unknown_provider_name!r} not found"
    providers_for_overriding = {unknown_provider_name: None}

    with (
        pytest.raises(RuntimeError, match=match),
        container.DIContainer.override_providers_sync(providers_for_overriding),
    ):
        ...  # pragma: no cover


async def test_providers_overriding() -> None:
    async_resource_mock = datetime.datetime.fromisoformat("2023-01-01")
    sync_resource_mock = datetime.datetime.fromisoformat("2024-01-01")
    async_factory_mock = datetime.datetime.fromisoformat("2025-01-01")
    simple_factory_mock = container.SimpleFactory(dep1="override", dep2=999)
    singleton_mock = container.SingletonFactory(dep1=False)
    object_mock = object()
    container.DIContainer.async_resource.override_sync(async_resource_mock)
    container.DIContainer.sync_resource.override_sync(sync_resource_mock)
    container.DIContainer.simple_factory.override_sync(simple_factory_mock)
    container.DIContainer.singleton.override_sync(singleton_mock)
    container.DIContainer.async_factory.override_sync(async_factory_mock)
    container.DIContainer.object.override_sync(object_mock)

    await container.DIContainer.simple_factory()
    dependent_factory = await container.DIContainer.dependent_factory()
    singleton = await container.DIContainer.singleton()
    async_factory = await container.DIContainer.async_factory()
    obj = await container.DIContainer.object()

    assert dependent_factory.simple_factory.dep1 == simple_factory_mock.dep1
    assert dependent_factory.simple_factory.dep2 == simple_factory_mock.dep2
    assert dependent_factory.sync_resource == sync_resource_mock
    assert dependent_factory.async_resource == async_resource_mock
    assert singleton is singleton_mock
    assert async_factory is async_factory_mock
    assert obj is object_mock

    container.DIContainer.reset_override_sync()
    assert (await container.DIContainer.async_resource()) != async_resource_mock


async def test_providers_overriding_sync_resolve() -> None:
    async_resource_mock = datetime.datetime.fromisoformat("2023-01-01")
    sync_resource_mock = datetime.datetime.fromisoformat("2024-01-01")
    simple_factory_mock = container.SimpleFactory(dep1="override", dep2=999)
    singleton_mock = container.SingletonFactory(dep1=False)
    object_mock = object()
    container.DIContainer.async_resource.override_sync(async_resource_mock)
    container.DIContainer.sync_resource.override_sync(sync_resource_mock)
    container.DIContainer.simple_factory.override_sync(simple_factory_mock)
    container.DIContainer.singleton.override_sync(singleton_mock)
    container.DIContainer.object.override_sync(object_mock)

    container.DIContainer.simple_factory.resolve_sync()
    await container.DIContainer.async_resource.resolve()
    dependent_factory = container.DIContainer.dependent_factory.resolve_sync()
    singleton = container.DIContainer.singleton.resolve_sync()
    obj = container.DIContainer.object.resolve_sync()

    assert dependent_factory.simple_factory.dep1 == simple_factory_mock.dep1
    assert dependent_factory.simple_factory.dep2 == simple_factory_mock.dep2
    assert dependent_factory.sync_resource == sync_resource_mock
    assert dependent_factory.async_resource == async_resource_mock
    assert singleton is singleton_mock
    assert obj is object_mock

    container.DIContainer.reset_override_sync()
    assert container.DIContainer.sync_resource.resolve_sync() != sync_resource_mock


async def test_provider_tear_down_after_override() -> None:
    original_value = 100
    override_value = 32

    class _MyContainer(BaseContainer):
        B = providers.Singleton(lambda: original_value)
        A = providers.Singleton(lambda x: x, B)

    a_old = await _MyContainer.A.resolve()

    _MyContainer.B.override_sync(override_value, tear_down_children=True)

    a_new = await _MyContainer.A()

    assert a_old != a_new

    _MyContainer.B.reset_override_sync(tear_down_children=True)

    assert original_value == _MyContainer.B.resolve_sync()


async def test_provider_tear_down_after_async_override() -> None:
    original_value = 100
    override_value = 32

    class _MyContainer(BaseContainer):
        B = providers.Singleton(lambda: original_value)
        A = providers.Singleton(lambda x: x, B)

    a_old = await _MyContainer.A.resolve()

    await _MyContainer.B.override(override_value, tear_down_children=True)

    a_new = await _MyContainer.A()

    assert a_old != a_new

    await _MyContainer.B.reset_override(tear_down_children=True)

    assert original_value == _MyContainer.B.resolve_sync()


async def test_provider_sync_override_raises_on_async_teardown() -> None:
    original_value = 100

    async def _async_creator(x: int) -> typing.AsyncIterator[int]:
        yield x

    class _MyContainer(BaseContainer):
        B = providers.Singleton(lambda: original_value)
        A = providers.Resource(_async_creator, B.cast)

    await _MyContainer.A.resolve()
    with pytest.raises(CannotTearDownSyncError):
        _MyContainer.B.override_sync(20, tear_down_children=True, raise_on_async=True)

    with pytest.warns(RuntimeWarning):
        _MyContainer.B.override_sync(20, tear_down_children=True, raise_on_async=False)



import asyncio
import random
import threading
import time
import typing
from concurrent.futures.thread import ThreadPoolExecutor

import pytest

from that_depends.providers import AsyncFactory, ThreadLocalSingleton


random.seed(23)


async def _async_factory() -> int:
    await asyncio.sleep(0.01)
    return threading.get_ident()


def _factory() -> int:
    time.sleep(0.01)
    return random.randint(1, 100)


def test_thread_local_singleton_same_thread() -> None:
    """Test that the same instance is returned within a single thread."""
    provider = ThreadLocalSingleton(_factory)

    instance1 = provider.resolve_sync()
    instance2 = provider.resolve_sync()

    assert instance1 == instance2, "Singleton failed: Instances within the same thread should be identical."

    provider.tear_down_sync()

    assert provider._instance is None, "Tear down failed: Instance should be reset to None."


async def test_async_thread_local_singleton_asyncio() -> None:
    """Test that the same instance is returned within a single thread."""
    provider = ThreadLocalSingleton(_factory)

    instance1 = await provider.resolve()
    instance2 = await provider.resolve()

    assert instance1 == instance2, "Singleton failed: Instances within the same thread should be identical."

    await provider.tear_down()

    assert provider._instance is None, "Tear down failed: Instance should be reset to None."


def test_thread_local_singleton_different_threads() -> None:
    """Test that different threads receive different instances."""
    provider = ThreadLocalSingleton(_factory)
    results = []

    def resolve_in_thread() -> None:
        results.append(provider.resolve_sync())

    number_of_threads = 10

    threads = [threading.Thread(target=resolve_in_thread) for _ in range(number_of_threads)]

    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()

    assert len(results) == number_of_threads, "Test failed: Expected results from two threads."
    assert results[0] != results[1], "Thread-local failed: Instances across threads should differ."


def test_thread_local_singleton_override() -> None:
    """Test overriding the ThreadLocalSingleton and resetting the override."""
    provider = ThreadLocalSingleton(_factory)

    override_value = 101
    provider.override_sync(override_value)
    instance = provider.resolve_sync()
    assert instance == override_value, "Override failed: Expected overridden value."

    # Reset override and ensure a new instance is created
    provider.reset_override_sync()
    new_instance = provider.resolve_sync()
    assert new_instance != override_value, "Reset override failed: Should no longer return overridden value."


def test_thread_local_singleton_override_in_threads() -> None:
    """Test that resetting the override in one thread does not affect another thread."""
    provider = ThreadLocalSingleton(_factory)
    results = {}

    def _thread_task(thread_id: int, override_value: int | None = None) -> None:
        if override_value is not None:
            provider.override_sync(override_value)
        results[thread_id] = provider.resolve_sync()
        if override_value is not None:
            provider.reset_override_sync()

    override_value: typing.Final[int] = 101
    thread1 = threading.Thread(target=_thread_task, args=(1, override_value))
    thread2 = threading.Thread(target=_thread_task, args=(2,))

    thread1.start()
    thread2.start()

    thread1.join()
    thread2.join()

    # Validate results
    assert results[1] == override_value, "Thread 1: Override failed."
    assert results[2] != override_value, "Thread 2: Should not be affected by Thread 1's override."
    assert results[1] != results[2], "Instances should be unique across threads."


def test_thread_local_singleton_override_temporarily() -> None:
    """Test temporarily overriding the ThreadLocalSingleton."""
    provider = ThreadLocalSingleton(_factory)

    override_value: typing.Final = 101
    # Set a temporary override
    with provider.override_context_sync(override_value):
        instance = provider.resolve_sync()
        assert instance == override_value, "Override context failed: Expected overridden value."

    # After the context, reset to the factory
    new_instance = provider.resolve_sync()
    assert new_instance != override_value, "Override context failed: Value should reset after context."


async def test_async_thread_local_singleton_asyncio_concurrency() -> None:
    singleton_async = ThreadLocalSingleton(_factory)

    expected = await singleton_async.resolve()

    results = await asyncio.gather(
        singleton_async(),
        singleton_async(),
        singleton_async(),
        singleton_async(),
        singleton_async(),
    )

    assert all(val is results[0] for val in results)
    for val in results:
        assert val == expected, "Instances should be identical across threads."


async def test_thread_local_singleton_async_resolve_with_async_dependencies() -> None:
    async_provider = AsyncFactory(_async_factory)

    def _dependent_creator(v: int) -> int:
        return v

    provider = ThreadLocalSingleton(_dependent_creator, v=async_provider.cast)

    expected = await provider.resolve()

    assert expected == await provider.resolve()

    results = await asyncio.gather(*[provider.resolve() for _ in range(10)])

    for val in results:
        assert val == expected, "Instances should be identical across threads."
    with pytest.raises(RuntimeError):
        # This should raise an error because the provider is async and resolution is attempted.
        await asyncio.gather(asyncio.to_thread(provider.resolve_sync))

    def _run_async_in_thread(coroutine: typing.Awaitable[typing.Any]) -> typing.Any:  # noqa: ANN401
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        result = loop.run_until_complete(coroutine)
        loop.close()
        return result

    with ThreadPoolExecutor() as executor:
        future = executor.submit(_run_async_in_thread, provider.resolve())

        result = future.result()

        assert result != expected, (
            "Since singleton should have been newly resolved, it should not have the same thread id."
        )


async def test_thread_local_singleton_async_resolve_override() -> None:
    provider = ThreadLocalSingleton(_factory)

    override_value = 101

    provider.override_sync(override_value)

    assert await provider.resolve() == override_value, "Override failed: Expected overridden value."



import datetime
import logging
import typing

import pydantic
import pytest

from tests.container import create_async_resource, create_sync_resource
from that_depends import BaseContainer, providers


logger = logging.getLogger(__name__)
global_state_for_selector: typing.Literal["sync_resource", "async_resource", "missing"] = "sync_resource"


class SelectorState:
    def __init__(self) -> None:
        self.selector_state: typing.Literal["sync_resource", "async_resource", "missing"] = "sync_resource"

    def get_selector_state(self) -> typing.Literal["sync_resource", "async_resource", "missing"]:
        return self.selector_state


selector_state = SelectorState()


class DIContainer(BaseContainer):
    alias = "selector_container"
    sync_resource = providers.Resource(create_sync_resource)
    async_resource = providers.Resource(create_async_resource)
    selector: providers.Selector[datetime.datetime] = providers.Selector(
        selector_state.get_selector_state,
        sync_resource=sync_resource,
        async_resource=async_resource,
    )


async def test_selector_provider_async() -> None:
    selector_state.selector_state = "async_resource"
    selected = await DIContainer.selector()
    async_resource = await DIContainer.async_resource()

    assert selected == async_resource


async def test_selector_provider_async_missing() -> None:
    selector_state.selector_state = "missing"
    with pytest.raises(KeyError, match="Key 'missing' not found in providers"):
        await DIContainer.selector()


async def test_selector_provider_sync() -> None:
    selector_state.selector_state = "sync_resource"
    selected = DIContainer.selector.resolve_sync()
    sync_resource = DIContainer.sync_resource.resolve_sync()

    assert selected == sync_resource


async def test_selector_provider_sync_missing() -> None:
    selector_state.selector_state = "missing"
    with pytest.raises(KeyError, match="Key 'missing' not found in providers"):
        DIContainer.selector.resolve_sync()


async def test_selector_provider_overriding() -> None:
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    DIContainer.selector.override_sync(now)
    selected_async = await DIContainer.selector()
    selected_sync = DIContainer.selector.resolve_sync()
    assert selected_async == selected_sync == now

    DIContainer.reset_override_sync()
    selector_state.selector_state = "sync_resource"
    selected = DIContainer.selector.resolve_sync()
    sync_resource = DIContainer.sync_resource.resolve_sync()
    assert selected == sync_resource


class Settings(pydantic.BaseModel):
    mode: typing.Literal["one", "two"] = "one"


class SettingsSelectorContainer(BaseContainer):
    settings = providers.Singleton(Settings)
    operator = providers.Selector(
        settings.cast.mode,
        one=providers.Singleton(lambda: "Provider 1"),
        two=providers.Singleton(lambda: "Provider 2"),
    )


async def test_selector_with_attrgetter_selector_sync() -> None:
    assert SettingsSelectorContainer.settings.resolve_sync().mode == "one"
    assert SettingsSelectorContainer.operator.resolve_sync() == "Provider 1"

    SettingsSelectorContainer.settings.override_sync(Settings(mode="two"))
    assert SettingsSelectorContainer.settings.resolve_sync().mode == "two"
    assert SettingsSelectorContainer.operator.resolve_sync() == "Provider 2"
    SettingsSelectorContainer.settings.reset_override_sync()


async def test_selector_with_attrgetter_selector_async() -> None:
    assert (await SettingsSelectorContainer.settings.resolve()).mode == "one"
    assert (await SettingsSelectorContainer.operator.resolve()) == "Provider 1"

    SettingsSelectorContainer.settings.override_sync(Settings(mode="two"))
    assert (await SettingsSelectorContainer.settings.resolve()).mode == "two"
    assert (await SettingsSelectorContainer.operator.resolve()) == "Provider 2"
    SettingsSelectorContainer.settings.reset_override_sync()


class StringSelectorContainer(BaseContainer):
    selector = providers.Selector(
        "one",
        one=providers.Singleton(lambda: "Provider 1"),
        two=providers.Singleton(lambda: "Provider 2"),
    )


async def test_selector_with_fixed_string() -> None:
    assert StringSelectorContainer.selector.resolve_sync() == "Provider 1"
    assert (await StringSelectorContainer.selector.resolve()) == "Provider 1"


async def mode_one() -> str:
    return "one"


class StringProviderSelectorContainer(BaseContainer):
    mode = providers.AsyncFactory(mode_one)
    selector = providers.Selector(
        mode.cast,
        one=providers.Singleton(lambda: "Provider 1"),
        two=providers.Singleton(lambda: "Provider 2"),
    )


async def test_selector_with_provider_selector_async() -> None:
    assert (await StringProviderSelectorContainer.selector.resolve()) == "Provider 1"


class InvalidSelectorContainer(BaseContainer):
    selector = providers.Selector(
        None,  # type: ignore[arg-type]
        one=providers.Singleton(lambda: "Provider 1"),
        two=providers.Singleton(lambda: "Provider 2"),
    )


async def test_selector_with_invalid_selector() -> None:
    with pytest.raises(
        TypeError, match="Invalid selector type: <class 'NoneType'>, expected str, or a provider/callable returning str"
    ):
        InvalidSelectorContainer.selector.resolve_sync()

    with pytest.raises(
        TypeError, match="Invalid selector type: <class 'NoneType'>, expected str, or a provider/callable returning str"
    ):
        await InvalidSelectorContainer.selector.resolve()



from that_depends import ContextScopes
from that_depends.meta import BaseContainerMeta


def test_base_container_meta_has_correct_default_scope() -> None:
    class _Test(metaclass=BaseContainerMeta):
        pass

    assert _Test.get_scope() == ContextScopes.ANY


def test_base_container_meta_has_correct_default_name() -> None:
    class _Test(metaclass=BaseContainerMeta):
        pass

    assert _Test.name() == "_Test"



import datetime

import pytest

from tests import container
from that_depends import BaseContainer, providers


class InnerContainer(BaseContainer):
    sync_resource = providers.Resource(container.create_sync_resource)
    async_resource = providers.Resource(container.create_async_resource)


class OuterContainer(BaseContainer):
    sequence = providers.List(InnerContainer.sync_resource, InnerContainer.async_resource)


OuterContainer.connect_containers(InnerContainer)


async def test_included_container() -> None:
    sequence = await OuterContainer.sequence()
    assert all(isinstance(x, datetime.datetime) for x in sequence)

    await OuterContainer.tear_down()
    assert InnerContainer.sync_resource._context.instance is None
    assert InnerContainer.async_resource._context.instance is None

    await OuterContainer.init_resources()
    sync_resource_context = InnerContainer.sync_resource._context
    assert sync_resource_context
    assert sync_resource_context.instance is not None
    async_resource_context = InnerContainer.async_resource._context
    assert async_resource_context
    assert async_resource_context.instance is not None
    await OuterContainer.tear_down()


async def test_overwriting_container_warns(recwarn: None) -> None:  # noqa:ARG001
    class _A(BaseContainer):
        pass

    with pytest.warns(UserWarning):

        class _A(BaseContainer):  # type: ignore[no-redef]
            pass

    class _A(BaseContainer):  # type: ignore[no-redef]
        pass


async def test_overwriting_container_with_alias_warns(recwarn: None) -> None:  # noqa:ARG001
    class _A(BaseContainer):
        alias = "a"

    with pytest.warns(UserWarning):

        class _A(BaseContainer):  # type: ignore[no-redef]
            alias = "a"

    class _A(BaseContainer):  # type: ignore[no-redef]
        alias = "a"



import asyncio
import functools
import inspect
import re
import threading
import typing
import warnings
from contextlib import AsyncExitStack, ExitStack

from that_depends.container import BaseContainer
from that_depends.meta import BaseContainerMeta
from that_depends.providers import AbstractProvider, ContextResource
from that_depends.providers.context_resources import ContextScope, ContextScopes


P = typing.ParamSpec("P")
T = typing.TypeVar("T")
_STRING_PROVIDER_PATTERN = re.compile(r"^([^.]+)\.([^.]+)(?:\.(.+))?$")


@typing.overload
def inject(func: typing.Callable[P, T]) -> typing.Callable[P, T]: ...


@typing.overload
def inject(
    *,
    scope: ContextScope | None = ContextScopes.INJECT,
) -> typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]: ...


def inject(
    func: typing.Callable[P, T] | None = None, scope: ContextScope | None = ContextScopes.INJECT
) -> typing.Callable[P, T] | typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]:
    """Inject dependencies into a function."""
    if scope == ContextScopes.ANY:
        msg = f"{scope} is not allowed in inject decorator."
        raise ValueError(msg)

    def _inject(
        func: typing.Callable[P, T],
    ) -> typing.Callable[P, T]:
        if inspect.iscoroutinefunction(func):
            return typing.cast(typing.Callable[P, T], _inject_to_async(func))

        return _inject_to_sync(func)

    def _inject_to_async(
        func: typing.Callable[P, typing.Coroutine[typing.Any, typing.Any, T]],
    ) -> typing.Callable[P, typing.Coroutine[typing.Any, typing.Any, T]]:
        @functools.wraps(func)
        async def inner(*args: P.args, **kwargs: P.kwargs) -> T:
            return await _resolve_async(func, scope, *args, **kwargs)

        return inner

    def _inject_to_sync(
        func: typing.Callable[P, T],
    ) -> typing.Callable[P, T]:
        @functools.wraps(func)
        def inner(*args: P.args, **kwargs: P.kwargs) -> T:
            return _resolve_sync(func, scope, *args, **kwargs)

        return inner

    if func:
        return _inject(func)

    return _inject


_SYNC_SIGNATURE_CACHE: dict[typing.Callable[..., typing.Any], inspect.Signature] = {}
_THREADING_LOCK = threading.Lock()


def _resolve_sync(
    func: typing.Callable[P, T],
    scope: ContextScope | None,
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    if func not in _SYNC_SIGNATURE_CACHE:
        with _THREADING_LOCK:
            _SYNC_SIGNATURE_CACHE[func] = inspect.signature(func)
    signature = _SYNC_SIGNATURE_CACHE[func]

    injected = False
    context_providers: set[AbstractProvider[typing.Any]] = set()

    with ExitStack() as stack:
        for i, (field_name, param) in enumerate(signature.parameters.items()):
            default = param.default

            if i < len(args) or field_name in kwargs:
                if isinstance(default, (AbstractProvider, StringProviderDefinition)):
                    injected = True
                continue

            if isinstance(default, StringProviderDefinition):
                injected = True
                kwargs[field_name] = _resolve_provider_with_scope_sync(
                    default.provider, scope, stack, context_providers
                )
            elif isinstance(default, AbstractProvider):
                injected = True
                kwargs[field_name] = _resolve_provider_with_scope_sync(default, scope, stack, context_providers)

        if not injected:
            warnings.warn(
                "Expected injection, but nothing found. Remove @inject decorator.", RuntimeWarning, stacklevel=1
            )

        return func(*args, **kwargs)


_SIGNATURE_CACHE: dict[
    typing.Callable[..., typing.Coroutine[typing.Any, typing.Any, typing.Any]], inspect.Signature
] = {}

_ASYNCIO_LOCK = asyncio.Lock()


async def _resolve_async(
    func: typing.Callable[P, typing.Coroutine[typing.Any, typing.Any, T]],
    scope: ContextScope | None,
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    if func not in _SIGNATURE_CACHE:
        async with _ASYNCIO_LOCK:
            _SIGNATURE_CACHE[func] = inspect.signature(func)
    signature = _SIGNATURE_CACHE[func]

    injected = False
    context_providers: set[AbstractProvider[typing.Any]] = set()

    async with AsyncExitStack() as stack:
        params = list(signature.parameters.items())

        for i, (field_name, param) in enumerate(params):
            default = param.default

            if i < len(args) or field_name in kwargs:
                if isinstance(default, (AbstractProvider, StringProviderDefinition)):
                    injected = True
                continue

            if isinstance(default, StringProviderDefinition):
                injected = True
                resolved_val = await _resolve_provider_with_scope_async(
                    default.provider, scope, stack, context_providers
                )
                kwargs[field_name] = resolved_val
            elif isinstance(default, AbstractProvider):
                injected = True
                resolved_val = await _resolve_provider_with_scope_async(default, scope, stack, context_providers)
                kwargs[field_name] = resolved_val

        if not injected:
            warnings.warn(
                "Expected injection, but nothing found. Remove @inject decorator.", RuntimeWarning, stacklevel=1
            )

        return await func(*args, **kwargs)

    raise RuntimeError  # pragma: no cover # to prevent mypy issue


async def _resolve_provider_with_scope_async(
    provider: AbstractProvider[T],
    scope: ContextScope | None,
    stack: AsyncExitStack,
    providers: set[AbstractProvider[typing.Any]],
) -> T:
    await _add_provider_to_stack_async(provider, stack, scope, providers)
    return await provider.resolve()


async def _add_provider_to_stack_async(
    provider: AbstractProvider[T],
    stack: AsyncExitStack,
    scope: ContextScope | None,
    providers: set[AbstractProvider[typing.Any]],
) -> None:
    if provider in providers:
        return
    providers.add(provider)

    if not (scope and isinstance(provider, ContextResource)):
        return

    provider_scope = provider.get_scope()

    if provider_scope not in (ContextScopes.ANY, scope):
        return
    provider._register_arguments()  # noqa: SLF001

    parents = provider._parents  # noqa: SLF001
    for parent in parents:
        await _add_provider_to_stack_async(parent, stack, scope, providers)
    await stack.enter_async_context(provider.context_async(force=True))


def _resolve_provider_with_scope_sync(
    provider: AbstractProvider[T],
    scope: ContextScope | None,
    stack: ExitStack,
    providers: set[AbstractProvider[typing.Any]],
) -> T:
    _add_provider_to_stack_sync(provider, stack, scope, providers)
    return provider.resolve_sync()


def _add_provider_to_stack_sync(
    provider: AbstractProvider[T],
    stack: ExitStack,
    scope: ContextScope | None,
    providers: set[AbstractProvider[typing.Any]],
) -> None:
    if provider in providers:
        return
    providers.add(provider)

    if not (scope and isinstance(provider, ContextResource)):
        return

    provider_scope = provider.get_scope()
    if provider_scope not in (ContextScopes.ANY, scope):
        return

    provider._register_arguments()  # noqa: SLF001

    parents = provider._parents  # noqa: SLF001
    for parent in parents:
        _add_provider_to_stack_sync(parent, stack, scope, providers)

    stack.enter_context(provider.context_sync(force=True))


class StringProviderDefinition:
    """Provider definition from a string."""

    def __init__(self, definition: str) -> None:
        """Initialize the provider definition.

        Args:
            definition: provider definition in the format container.provider[.attr1.attr2...].

        """
        self._definition = definition
        self._container_name, self._provider_name, self._attrs = self._validate_and_extract_provider_definition()

    def _validate_and_extract_provider_definition(self) -> tuple[str, str, list[str]]:
        match = re.match(_STRING_PROVIDER_PATTERN, self._definition)
        if match:
            container_name = match.group(1)
            provider_name = match.group(2)
            attrs = match.group(3).split(".") if match.group(3) else []
            return container_name, provider_name, attrs
        msg = f"Invalid provider definition: {self._definition}"
        raise ValueError(msg)

    def _get_container_by_name(self) -> type["BaseContainer"]:
        containers_in_scope = BaseContainerMeta.get_instances()
        try:
            return containers_in_scope[self._container_name]
        except KeyError as e:
            msg = f"Container {self._container_name} not found in scope!"
            raise ValueError(msg) from e

    def _get_provider_by_name(self) -> AbstractProvider[typing.Any]:
        container = self._get_container_by_name()
        try:
            provider = container.get_providers()[self._provider_name]
        except KeyError as e:
            msg = f"Provider {self._provider_name} not found in container {self._container_name}"
            raise ValueError(msg) from e
        for attr in self._attrs:
            provider = getattr(provider, attr)
        return provider

    @property
    def provider(self) -> AbstractProvider[typing.Any]:
        """Get the provider instance."""
        return self._get_provider_by_name()


class ClassGetItemMeta(type):
    """Metaclass to support Provide[provider] syntax."""

    def __getitem__(cls, provider: AbstractProvider[T] | str) -> T | typing.Any:  # noqa: ANN401
        if isinstance(provider, str):
            return StringProviderDefinition(provider)  # will be resolved later
        return typing.cast(T, provider)


class Provide(metaclass=ClassGetItemMeta):
    """Marker to dependency injection."""



import abc
import asyncio
import contextlib
import inspect
import logging
import typing
from abc import abstractmethod
from collections.abc import Iterable
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from contextvars import ContextVar, Token
from functools import wraps
from types import TracebackType
from typing import Final, overload

from typing_extensions import TypeIs, override

from that_depends.entities.resource_context import ResourceContext
from that_depends.providers.base import AbstractResource


if typing.TYPE_CHECKING:
    from that_depends.container import BaseContainer


logger: typing.Final = logging.getLogger(__name__)
T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")
_CONTAINER_CONTEXT: typing.Final[ContextVar[dict[str, typing.Any]]] = ContextVar("__CONTAINER_CONTEXT__")

AppType = typing.TypeVar("AppType")
Scope = typing.MutableMapping[str, typing.Any]
Message = typing.MutableMapping[str, typing.Any]
Receive = typing.Callable[[], typing.Awaitable[Message]]
Send = typing.Callable[[Message], typing.Awaitable[None]]
ASGIApp = typing.Callable[[Scope, Receive, Send], typing.Awaitable[None]]
_ASYNC_CONTEXT_KEY: typing.Final[str] = "__ASYNC_CONTEXT__"

ContextType = dict[str, typing.Any]


class InvalidContextError(RuntimeError):
    """Raised when an invalid context is being used."""


class ContextScope:
    """A named context scope."""

    def __init__(self, name: str) -> None:
        """Initialize a new context scope."""
        self._name = name

    @property
    def name(self) -> str:
        """Get the name of the context scope."""
        return self._name

    @override
    def __eq__(self, other: object) -> bool:
        if isinstance(other, ContextScope):
            return self.name == other.name
        return False

    @override
    def __repr__(self) -> str:
        return f"{self.name!r}"


class ContextScopes:
    """Enumeration of context scopes."""

    ANY = ContextScope("ANY")  # special scope that can be used in any context
    APP = ContextScope("APP")  # application scope
    REQUEST = ContextScope("REQUEST")  # request scope
    INJECT = ContextScope("INJECT")  # inject scope


_CONTAINER_SCOPE: typing.Final[ContextVar[ContextScope | None]] = ContextVar("__CONTAINER_SCOPE__", default=None)


def get_current_scope() -> ContextScope | None:
    """Get the current context scope.

    Returns:
        ContextScope | None: The current context scope.

    """
    return _CONTAINER_SCOPE.get()


def _set_current_scope(scope: ContextScope | None) -> Token[ContextScope | None]:
    return _CONTAINER_SCOPE.set(scope)


@contextlib.contextmanager
def _enter_named_scope(scope: ContextScope) -> typing.Iterator[ContextScope]:
    token = _set_current_scope(scope)
    yield scope
    _CONTAINER_SCOPE.reset(token)


T = typing.TypeVar("T")
CT = typing.TypeVar("CT")


class SupportsContext(typing.Generic[CT], abc.ABC):
    """Interface for resources that support context initialization.

    This interface defines methods to create synchronous and asynchronous
    context managers, as well as a function decorator for context initialization.
    """

    @abstractmethod
    def get_scope(self) -> ContextScope | None:
        """Return the scope of the resource."""

    @abstractmethod
    def context_async(self, force: bool = False) -> typing.AsyncContextManager[CT]:
        """Create an async context manager for this resource.

        Args:
            force (bool): If True, the context will be entered regardless of the current scope.

        Returns:
            AsyncContextManager[CT]: An async context manager.

        Example:
            ```python
            async with my_resource.context_async():
                result = await my_resource.resolve()
            ```

        """

    @abstractmethod
    def context_sync(self, force: bool = False) -> typing.ContextManager[CT]:
        """Create a sync context manager for this resource.

        Args:
            force (bool): If True, the context will be entered regardless of the current scope.

        Returns:
            ContextManager[CT]: A sync context manager.

        Example:
            ```python
            with my_resource.context_sync():
                result = my_resource.resolve_sync()
            ```

        """

    @abstractmethod
    def supports_context_sync(self) -> bool:
        """Check whether the resource supports sync context.

        Returns:
            bool: True if sync context is supported, False otherwise.

        """


def _get_container_context() -> dict[str, typing.Any] | None:
    try:
        return _CONTAINER_CONTEXT.get()
    except LookupError:
        return None


def fetch_context_item(key: str, default: typing.Any = None, raise_on_not_found: bool = False) -> typing.Any:  # noqa: ANN401
    """Retrieve a value from the global context.

    Args:
        key (str): The key to retrieve from the global context.
        default (Any): The default value to return if the key is not found.
        raise_on_not_found (bool): If True, raises a KeyError if the key is not found.

    Returns:
        Any: The value associated with the key in the global context or the default value.

    Example:
        ```python
        async with container_context(global_context={"username": "john_doe"}):
            user = fetch_context_item("username")
        ```

    """
    if context := _get_container_context():
        return context.get(key, default)
    if raise_on_not_found:
        msg = f"Key `{key}` not found in global context."
        raise KeyError(msg)
    return default


def fetch_context_item_by_type(t: type[T]) -> T | None:
    """Retrieve a value from the global context by type.

    Args:
        t (type[T]): The type of the value to retrieve.

    Returns:
        T | None: The value associated with the type in the global context or None if not found.

    Raises:
        RuntimeError: If the context item of the specified type is not found in the global context.

    """
    if context := _get_container_context():
        for value in context.values():
            if isinstance(value, t):
                return value
    msg = f"Cannot find context item of type {t} in the global context."
    raise RuntimeError(msg)


class ContextResource(
    AbstractResource[T_co],
    SupportsContext[ResourceContext[T_co]],
):
    """A context-dependent provider that resolves resources only if their context is initialized.

    `ContextResource` handles both synchronous and asynchronous resource creators
    and ensures they are properly torn down when the context exits.
    """

    @override
    async def resolve(self) -> T_co:
        current_scope = get_current_scope()
        if not self._strict_scope or self._scope in (ContextScopes.ANY, current_scope):
            return await super().resolve()
        msg = f"Cannot resolve resource with scope `{self._scope}` in scope `{current_scope}`"
        raise RuntimeError(msg)

    @override
    def resolve_sync(self) -> T_co:
        current_scope = get_current_scope()
        if not self._strict_scope or self._scope in (ContextScopes.ANY, current_scope):
            return super().resolve_sync()
        msg = f"Cannot resolve resource with scope `{self._scope}` in scope `{current_scope}`"
        raise RuntimeError(msg)

    @override
    def get_scope(self) -> ContextScope | None:
        return self._scope

    __slots__ = (
        "_args",
        "_context",
        "_creator",
        "_internal_name",
        "_is_async",
        "_kwargs",
        "_override",
        "_scope",
        "_token",
    )

    def __init__(
        self,
        creator: typing.Callable[P, typing.Iterator[T_co] | typing.AsyncIterator[T_co]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        """Initialize a new context resource.

        Args:
            creator (Callable[P, Iterator[T_co] | AsyncIterator[T_co]]):
                A sync or async iterator that yields the resource to be provided.
            *args (P.args): Positional arguments to pass to the creator.
            **kwargs (P.kwargs): Keyword arguments to pass to the creator.

        """
        super().__init__(creator, *args, **kwargs)
        self._from_creator = creator
        self._context: ContextVar[ResourceContext[T_co]] = ContextVar(f"{self._creator.__name__}-context")
        self._token: Token[ResourceContext[T_co]] | None = None
        self._async_lock: Final = asyncio.Lock()
        self._scope: ContextScope | None = ContextScopes.ANY
        self._strict_scope: bool = False

    @overload
    def context(self, func: typing.Callable[P, T]) -> typing.Callable[P, T]: ...

    @overload
    def context(self, *, force: bool = False) -> typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]: ...

    def context(
        self, func: typing.Callable[P, T] | None = None, force: bool = False
    ) -> typing.Callable[P, T] | typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]:
        """Create a new context manager for the resource, the context manager will be async if the resource is async.

        Returns:
            typing.ContextManager[ResourceContext[T_co]] | typing.AsyncContextManager[ResourceContext[T_co]]:
            A context manager for the resource.

        """

        def _wrapper(func: typing.Callable[P, T]) -> typing.Callable[P, T]:
            if inspect.iscoroutinefunction(func):

                @wraps(func)
                async def _async_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                    async with self.context_async(force=force):
                        return await func(*args, **kwargs)  # type: ignore[no-any-return, misc]

                return typing.cast(typing.Callable[P, T], _async_wrapper)

            # wrapped function is sync
            @wraps(func)
            def _sync_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                with self.context_sync(force=force):
                    return func(*args, **kwargs)

            return typing.cast(typing.Callable[P, T], _sync_wrapper)

        if func:
            return _wrapper(func)
        return _wrapper

    def with_config(self, scope: ContextScope | None, strict_scope: bool = False) -> "ContextResource[T_co]":
        """Create a new context-resource with the specified scope.

        Args:
            scope: named scope where resource is resolvable.
            strict_scope: if True, the resource will only be resolvable in the specified scope.

        Returns:
            new context resource with the specified scope.

        """
        if strict_scope and scope == ContextScopes.ANY:
            msg = f"Cannot set strict_scope with scope {scope}."
            raise ValueError(msg)
        r = ContextResource(self._from_creator, *self._args, **self._kwargs)  # type: ignore[arg-type]
        r._scope = scope
        r._strict_scope = strict_scope

        return r

    @override
    def supports_context_sync(self) -> bool:
        return not self._is_async

    def _enter_context_sync(self, force: bool = False) -> ResourceContext[T_co]:
        if self._is_async:
            msg = "You must enter async context for async creators."
            raise RuntimeError(msg)
        return self._enter(force)

    async def _enter_context_async(self, force: bool = False) -> ResourceContext[T_co]:
        return self._enter(force)

    def _enter(self, force: bool = False) -> ResourceContext[T_co]:
        if not force and self._scope not in (ContextScopes.ANY, get_current_scope()):
            msg = f"Cannot enter context for resource with scope {self._scope} in scope {get_current_scope()!r}"
            raise InvalidContextError(msg)
        self._token = self._context.set(ResourceContext(is_async=self._is_async))
        return self._context.get()

    def _exit_context_sync(self) -> None:
        if not self._token:
            msg = "Context is not set, call ``_enter_sync_context`` first"
            raise RuntimeError(msg)

        try:
            context_item = self._context.get()
            context_item.tear_down_sync()
        finally:
            self._context.reset(self._token)

    async def _exit_context_async(self) -> None:
        if self._token is None:
            msg = "Context is not set, call ``_enter_async_context`` first"
            raise RuntimeError(msg)

        try:
            context_item = self._context.get()
            if context_item.is_context_stack_async(context_item.context_stack):
                await context_item.tear_down()
            else:
                context_item.tear_down_sync()
        finally:
            self._context.reset(self._token)

    @contextlib.contextmanager
    @override
    def context_sync(self, force: bool = False) -> typing.Iterator[ResourceContext[T_co]]:
        if self._is_async:
            msg = "Please use async context instead."
            raise RuntimeError(msg)
        token = self._token
        with self._lock:
            val = self._enter_context_sync(force=force)
            temp_token = self._token
        yield val
        with self._lock:
            self._token = temp_token
            self._exit_context_sync()
        self._token = token

    @contextlib.asynccontextmanager
    @override
    async def context_async(self, force: bool = False) -> typing.AsyncIterator[ResourceContext[T_co]]:
        token = self._token

        async with self._async_lock:
            val = await self._enter_context_async(force=force)
            temp_token = self._token
        yield val
        async with self._async_lock:
            self._token = temp_token
            await self._exit_context_async()
        self._token = token

    def _fetch_context(self) -> ResourceContext[T_co]:
        try:
            return self._context.get()
        except LookupError as e:
            msg = "Context is not set. Use container_context"
            raise RuntimeError(msg) from e


ContainerType = typing.TypeVar("ContainerType", bound="type[BaseContainer]")


class container_context(AbstractContextManager[ContextType], AbstractAsyncContextManager[ContextType]):  # noqa: N801
    """Initialize contexts for the provided containers or resources.

    Use this class to manage global and resource-specific contexts in both
    synchronous and asynchronous scenarios.
    """

    ___slots__ = (
        "_providers",
        "_context_stack",
        "_containers",
        "_initial_context",
        "_context_token",
        "_scope",
    )

    def __init__(
        self,
        *context_items: SupportsContext[typing.Any],
        global_context: ContextType | None = None,
        preserve_global_context: bool = True,
        scope: ContextScope | None = None,
    ) -> None:
        """Initialize a new container context.

        Args:
            *context_items (SupportsContext[Any]): Context items to initialize a new context for.
            global_context (dict[str, Any] | None): A dictionary representing the global context.
            preserve_global_context (bool): If True, merges the existing global context with the new one.
            scope (ContextScope | None): The named scope that should be initialized.

        Example:
            ```python
            async with container_context(MyContainer, global_context={"key": "value"}):
                data = fetch_context_item("key")
            ```

        """
        if scope == ContextScopes.ANY:
            msg = f"{scope} cannot be entered!"
            raise ValueError(msg)
        if len(context_items) == 0 and not scope and not global_context:
            msg = "One of context_items, scope or global_context must be provided."
            raise ValueError(msg)
        self._scope = scope
        self._preserve_global_context = preserve_global_context
        self._global_context = global_context
        self._context_token: Token[ContextType] | None = None
        self._context_items: typing.Final[set[SupportsContext[typing.Any]]] = set(context_items)
        self._context_providers: set[ContextResource[typing.Any]] = set()
        self._reset_resource_context: typing.Final[bool] = bool(scope)
        self._context_stack: contextlib.AsyncExitStack | contextlib.ExitStack | None = None
        self._scope_token: Token[ContextScope | None] | None = None

    def _resolve_initial_conditions(self) -> None:
        self._scope = self._scope if self._scope else get_current_scope()
        if self._preserve_global_context and self._global_context:
            if context := _get_container_context():
                self._initial_context = {**context, **self._global_context}
            else:
                self._initial_context = self._global_context
        elif context := _get_container_context():
            self._initial_context: ContextType = (  # type: ignore[no-redef]
                context if self._preserve_global_context else self._global_context or {}
            )
        else:
            self._initial_context = self._global_context or {}
        if self._reset_resource_context:
            from that_depends.meta import BaseContainerMeta

            self._add_providers_from_containers(BaseContainerMeta.get_instances().values(), self._scope)
        for item in self._context_items:
            from that_depends.container import BaseContainer

            if isinstance(item, type) and issubclass(item, BaseContainer):
                self._add_providers_from_containers([item], self._scope)
            elif isinstance(item, ContextResource):
                self._context_providers.add(item)

    def _add_providers_from_containers(
        self, containers: Iterable[ContainerType], scope: ContextScope | None = ContextScopes.ANY
    ) -> None:
        for container in containers:
            for container_provider in container.get_providers().values():
                if isinstance(container_provider, ContextResource):
                    provider_scope = container_provider.get_scope()
                    if provider_scope in (scope, ContextScopes.ANY):
                        self._context_providers.add(container_provider)

    @override
    def __enter__(self) -> ContextType:
        self._resolve_initial_conditions()
        self._context_stack = contextlib.ExitStack()
        self._scope_token = _set_current_scope(self._scope)
        for item in self._context_providers:
            if item.supports_context_sync():
                self._context_stack.enter_context(item.context_sync())
        return self._enter_globals()

    @override
    async def __aenter__(self) -> ContextType:
        self._resolve_initial_conditions()
        self._context_stack = contextlib.AsyncExitStack()
        self._scope_token = _set_current_scope(self._scope)
        for item in self._context_providers:
            await self._context_stack.enter_async_context(item.context_async())
        return self._enter_globals()

    def _enter_globals(self) -> ContextType:
        self._context_token = _CONTAINER_CONTEXT.set(self._initial_context)
        return _CONTAINER_CONTEXT.get()

    def _is_context_token(self, _: Token[ContextType] | None) -> TypeIs[Token[ContextType]]:
        return _ is not None

    def _is_scope_token(self, _: Token[ContextScope | None] | None) -> TypeIs[Token[ContextScope | None]]:
        return _ is not None

    def _exit_globals(self) -> None:
        if self._is_context_token(self._context_token):
            _CONTAINER_CONTEXT.reset(self._context_token)
        else:
            msg = "No context token set for global vars, use __enter__ or __aenter__ first."
            raise RuntimeError(msg)
        if self._is_scope_token(self._scope_token):
            _CONTAINER_SCOPE.reset(self._scope_token)

    def _has_async_exit_stack(
        self,
        _: contextlib.AsyncExitStack | contextlib.ExitStack | None,
    ) -> typing.TypeGuard[contextlib.AsyncExitStack]:
        return isinstance(_, contextlib.AsyncExitStack)

    def _has_sync_exit_stack(
        self, _: contextlib.AsyncExitStack | contextlib.ExitStack | None
    ) -> typing.TypeGuard[contextlib.ExitStack]:
        return isinstance(_, contextlib.ExitStack)

    @override
    def __exit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        try:
            if self._has_sync_exit_stack(self._context_stack):
                self._context_stack.close()
            else:
                msg = "Context is not set, call ``__enter__`` first"
                raise RuntimeError(msg)
        finally:
            self._exit_globals()

    @override
    async def __aexit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, traceback: TracebackType | None
    ) -> None:
        try:
            if self._has_async_exit_stack(self._context_stack):
                await self._context_stack.aclose()
            else:
                msg = "Context is not set, call ``__aenter__`` first"
                raise RuntimeError(msg)
        finally:
            self._exit_globals()

    def __call__(self, func: typing.Callable[P, T_co]) -> typing.Callable[P, T_co]:
        """Decorate a function to run within this container context.

        The context is automatically initialized before the function is called and
        torn down afterward.

        Args:
            func (Callable[P, T_co]): A sync or async callable.

        Returns:
            Callable[P, T_co]: The wrapped function.

        Example:
            ```python
            @container_context(MyContainer)
            async def my_async_function():
                result = await MyContainer.some_resource.async_resolve()
                return result
            ```

        """
        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def _async_inner(*args: P.args, **kwargs: P.kwargs) -> T_co:
                async with container_context(
                    *self._context_items,
                    scope=self._scope,
                    global_context=self._global_context,
                    preserve_global_context=self._preserve_global_context,
                ):
                    return await func(*args, **kwargs)  # type: ignore[no-any-return]

            return typing.cast(typing.Callable[P, T_co], _async_inner)

        @wraps(func)
        def _sync_inner(*args: P.args, **kwargs: P.kwargs) -> T_co:
            with container_context(
                *self._context_items,
                scope=self._scope,
                global_context=self._global_context,
                preserve_global_context=self._preserve_global_context,
            ):
                return func(*args, **kwargs)

        return _sync_inner


class DIContextMiddleware:
    """ASGI middleware that manages context initialization for incoming requests.

    This middleware automatically creates and tears down context for each request,
    ensuring that resources defined in containers or as context items are properly
    initialized and cleaned up.
    """

    def __init__(
        self,
        app: ASGIApp,
        *context_items: SupportsContext[typing.Any],
        global_context: dict[str, typing.Any] | None = None,
        scope: ContextScope | None = None,
    ) -> None:
        """Initialize the DIContextMiddleware.

        Args:
            app (ASGIApp): The ASGI application to wrap.
            *context_items (SupportsContext[Any]): A collection of containers and providers that
                need context initialization prior to a request.
            global_context (dict[str, Any] | None): A global context dictionary to set before requests.
            scope (ContextScope | None): The scope in which the context should be initialized.

        Example:
            ```python
            my_app.add_middleware(DIContextMiddleware, MyContainer, global_context={"api_key": "secret"})
            ```

        """
        self.app: typing.Final = app
        self._context_items: set[SupportsContext[typing.Any]] = set(context_items)
        self._global_context: dict[str, typing.Any] | None = global_context
        if scope == ContextScopes.ANY:
            msg = f"{scope} cannot be entered!"
            raise ValueError(msg)
        self._scope = scope

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """Handle the incoming ASGI request by initializing and tearing down context.

        The context is initialized before the request is processed and
        closed after the request is completed.

        Args:
            scope (ContextScope): The ASGI scope.
            receive (Receive): The receive call.
            send (Send): The send call.

        Returns:
            None

        """
        async with (
            container_context(*self._context_items, global_context=self._global_context, scope=self._scope)
            if self._context_items
            else container_context(global_context=self._global_context, scope=self._scope)
        ):
            return await self.app(scope, receive, send)



import datetime
import typing

import fastapi
import pytest
from starlette import status
from starlette.testclient import TestClient

from tests import container
from that_depends import fetch_context_item
from that_depends.providers import DIContextMiddleware


_GLOBAL_CONTEXT: typing.Final[dict[str, str]] = {"test2": "value2", "test1": "value1"}


@pytest.fixture
def fastapi_app() -> fastapi.FastAPI:
    app = fastapi.FastAPI()
    app.add_middleware(DIContextMiddleware, container.DIContainer, global_context=_GLOBAL_CONTEXT)

    @app.get("/")
    async def read_root(
        dependency: typing.Annotated[
            container.DependentFactory,
            fastapi.Depends(container.DIContainer.dependent_factory),
        ],
        free_dependency: typing.Annotated[
            container.FreeFactory,
            fastapi.Depends(container.DIContainer.resolver(container.FreeFactory)),
        ],
        singleton: typing.Annotated[
            container.SingletonFactory,
            fastapi.Depends(container.DIContainer.singleton),
        ],
        singleton_attribute: typing.Annotated[bool, fastapi.Depends(container.DIContainer.singleton.dep1)],
        context_resource: typing.Annotated[datetime.datetime, fastapi.Depends(container.DIContainer.context_resource)],
    ) -> datetime.datetime:
        assert dependency.sync_resource == free_dependency.dependent_factory.sync_resource
        assert dependency.async_resource == free_dependency.dependent_factory.async_resource
        assert singleton.dep1 is True
        assert singleton_attribute is True
        assert context_resource == await container.DIContainer.context_resource.resolve()
        for key, value in _GLOBAL_CONTEXT.items():
            assert fetch_context_item(key) == value
        return dependency.async_resource

    return app


@pytest.fixture
def fastapi_client(fastapi_app: fastapi.FastAPI) -> TestClient:
    return TestClient(fastapi_app)


async def test_read_main(fastapi_client: TestClient) -> None:
    response = fastapi_client.get("/")
    assert response.status_code == status.HTTP_200_OK
    assert (
        datetime.datetime.fromisoformat(response.json().replace("Z", "+00:00"))
        == await container.DIContainer.async_resource()
    )



import random
import time
import typing

from that_depends import BaseContainer, ContextScopes, Provide, inject, providers


def _iterator(x: float) -> typing.Iterator[float]:
    yield x


class _Container(BaseContainer):
    grandparent = providers.Singleton(lambda: random.random())
    parent = providers.Factory(lambda x: x, grandparent.cast)
    item = providers.ContextResource(_iterator, parent.cast).with_config(scope=ContextScopes.REQUEST)


@inject(scope=ContextScopes.REQUEST)
def _injected(
    x_1: float = Provide[_Container.grandparent],
    x_2: float = Provide[_Container.parent],
    x_3: float = Provide[_Container.item],
) -> float:
    return x_1 + x_2 + x_3


def _bench(n_iterations: int) -> float:
    start = time.time()
    for _ in range(n_iterations):
        _injected()
    end = time.time()
    _Container.tear_down_sync()
    return end - start


if __name__ == "__main__":
    for n in [10000, 100000, 1000000]:
        duration = _bench(n)



import asyncio
import logging
import threading
import time
import typing
from concurrent.futures import ThreadPoolExecutor, as_completed

import pytest

from tests.creators import (
    AsyncContextManagerResource,
    ContextManagerResource,
    create_async_resource,
    create_sync_resource,
)
from that_depends import BaseContainer, providers


logger = logging.getLogger(__name__)


class DIContainer(BaseContainer):
    alias = "resources_container"
    async_resource = providers.Resource(create_async_resource)
    sync_resource = providers.Resource(create_sync_resource)
    async_resource_from_class = providers.Resource(AsyncContextManagerResource)
    sync_resource_from_class = providers.Resource(ContextManagerResource)


@pytest.fixture(autouse=True)
async def _tear_down() -> typing.AsyncIterator[None]:
    try:
        yield
    finally:
        await DIContainer.tear_down()


async def test_async_resource() -> None:
    async_resource1 = await DIContainer.async_resource.resolve()
    async_resource2 = DIContainer.async_resource.resolve_sync()
    assert async_resource1 is async_resource2


async def test_async_resource_from_class() -> None:
    async_resource1 = await DIContainer.async_resource_from_class.resolve()
    async_resource2 = DIContainer.async_resource_from_class.resolve_sync()
    assert async_resource1 is async_resource2


async def test_sync_resource() -> None:
    sync_resource1 = await DIContainer.sync_resource.resolve()
    sync_resource2 = await DIContainer.sync_resource.resolve()
    assert sync_resource1 is sync_resource2


async def test_sync_resource_from_class() -> None:
    sync_resource1 = await DIContainer.sync_resource_from_class.resolve()
    sync_resource2 = await DIContainer.sync_resource_from_class.resolve()
    assert sync_resource1 is sync_resource2


async def test_async_resource_overridden() -> None:
    async_resource1 = await DIContainer.sync_resource.resolve()

    DIContainer.sync_resource.override_sync("override")

    async_resource2 = DIContainer.sync_resource.resolve_sync()
    async_resource3 = await DIContainer.sync_resource.resolve()

    DIContainer.sync_resource.reset_override_sync()

    async_resource4 = DIContainer.sync_resource.resolve_sync()

    assert async_resource2 is not async_resource1
    assert async_resource2 is async_resource3
    assert async_resource4 is async_resource1


async def test_sync_resource_overridden() -> None:
    sync_resource1 = await DIContainer.sync_resource.resolve()

    DIContainer.sync_resource.override_sync("override")

    sync_resource2 = DIContainer.sync_resource.resolve_sync()
    sync_resource3 = await DIContainer.sync_resource.resolve()

    DIContainer.sync_resource.reset_override_sync()

    sync_resource4 = DIContainer.sync_resource.resolve_sync()

    assert sync_resource2 is not sync_resource1
    assert sync_resource2 is sync_resource3
    assert sync_resource4 is sync_resource1


async def test_resource_with_empty_list() -> None:
    def create_sync_resource_with_list() -> typing.Iterator[list[typing.Any]]:
        logger.debug("Resource initiated")
        yield []
        logger.debug("Resource destructed")

    async def create_async_resource_with_dict() -> typing.AsyncIterator[dict[str, typing.Any]]:
        logger.debug("Async resource initiated")
        yield {}
        logger.debug("Async resource destructed")

    sync_resource = providers.Resource(create_sync_resource_with_list)
    async_resource = providers.Resource(create_async_resource_with_dict)

    sync_resource1 = await sync_resource()
    sync_resource2 = await sync_resource()
    assert sync_resource1 is sync_resource2

    async_resource1 = await async_resource()
    async_resource2 = await async_resource()
    assert async_resource1 is async_resource2

    await sync_resource.tear_down()
    await async_resource.tear_down()


async def test_resource_unsupported_creator() -> None:
    with pytest.raises(TypeError, match="Unsupported resource type"):
        providers.Resource(None)  # type: ignore[arg-type]


@pytest.mark.repeat(10)
async def test_async_resource_asyncio_concurrency() -> None:
    calls: int = 0

    async def create_resource() -> typing.AsyncIterator[str]:
        nonlocal calls
        calls += 1
        await asyncio.sleep(0)
        yield ""

    resource = providers.Resource(create_resource)

    await asyncio.gather(resource.resolve(), resource.resolve())

    assert calls == 1


@pytest.mark.repeat(10)
async def test_async_resource_async_teardown() -> None:
    calls: int = 0

    async def _create_resource() -> typing.AsyncIterator[str]:
        yield ""
        nonlocal calls
        await asyncio.sleep(0.01)
        calls += 1

    resource = providers.Resource(_create_resource)

    await resource.resolve()

    await asyncio.gather(*[resource.tear_down() for _ in range(10)])

    assert calls == 1


@pytest.mark.repeat(10)
def test_resource_threading_concurrency() -> None:
    calls: int = 0
    lock = threading.Lock()

    def create_resource() -> typing.Iterator[str]:
        nonlocal calls
        with lock:
            calls += 1
        time.sleep(0.01)
        yield ""

    resource = providers.Resource(create_resource)

    with ThreadPoolExecutor(max_workers=4) as pool:
        tasks = [
            pool.submit(resource.resolve_sync),
            pool.submit(resource.resolve_sync),
            pool.submit(resource.resolve_sync),
            pool.submit(resource.resolve_sync),
        ]
        results = [x.result() for x in as_completed(tasks)]

    assert results == ["", "", "", ""]
    assert calls == 1


def test_sync_resource_sync_tear_down() -> None:
    DIContainer.sync_resource.resolve_sync()
    assert DIContainer.sync_resource._context.instance is not None
    DIContainer.sync_resource.tear_down_sync()
    assert DIContainer.sync_resource._context.instance is None


async def test_async_resource_sync_tear_down_raises() -> None:
    await DIContainer.async_resource.resolve()
    assert DIContainer.async_resource._context.instance is not None
    with pytest.raises(RuntimeError):
        DIContainer.async_resource.tear_down_sync()



import datetime

from litestar import Litestar, get
from litestar.di import Provide
from litestar.status_codes import HTTP_200_OK
from litestar.testing import TestClient

from tests import container


@get("/")
async def index(injected: datetime.datetime) -> datetime.datetime:
    return injected


app = Litestar([index], dependencies={"injected": Provide(container.DIContainer.async_resource)})


async def test_litestar_di() -> None:
    with TestClient(app=app) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK, response.text
        assert (
            datetime.datetime.fromisoformat(response.json().replace("Z", "+00:00"))
            == await container.DIContainer.async_resource()
        )



import asyncio
import dataclasses
import threading
import time
import typing
from concurrent.futures import ThreadPoolExecutor, as_completed

import pydantic
import pytest

from that_depends import BaseContainer, providers


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class SingletonFactory:
    dep1: str


class Settings(pydantic.BaseModel):
    some_setting: str = "some_value"
    other_setting: str = "other_value"


async def create_async_obj(value: str) -> SingletonFactory:
    await asyncio.sleep(0.001)
    return SingletonFactory(dep1=f"async {value}")


async def _async_creator() -> int:
    await asyncio.sleep(0.001)
    return threading.get_ident()


def _sync_creator_with_dependency(dep: int) -> str:
    return f"Singleton {dep}"


class DIContainer(BaseContainer):
    alias = "singleton_container"
    factory: providers.AsyncFactory[int] = providers.AsyncFactory(_async_creator)
    settings: Settings = providers.Singleton(Settings).cast
    singleton = providers.Singleton(SingletonFactory, dep1=settings.some_setting)
    singleton_async = providers.AsyncSingleton(create_async_obj, value=settings.some_setting)
    singleton_with_dependency = providers.Singleton(_sync_creator_with_dependency, dep=factory.cast)


async def test_singleton_provider() -> None:
    singleton1 = await DIContainer.singleton()
    singleton2 = await DIContainer.singleton()
    singleton3 = DIContainer.singleton.resolve_sync()
    await DIContainer.singleton.tear_down()
    singleton4 = DIContainer.singleton.resolve_sync()

    assert singleton1 is singleton2 is singleton3
    assert singleton4 is not singleton1

    await DIContainer.tear_down()


async def test_singleton_attr_getter() -> None:
    singleton1 = await DIContainer.singleton()

    assert singleton1.dep1 == Settings().some_setting

    await DIContainer.tear_down()


async def test_singleton_with_empty_list() -> None:
    singleton = providers.Singleton(list[object])

    singleton1 = await singleton()
    singleton2 = await singleton()
    assert singleton1 is singleton2


@pytest.mark.repeat(10)
async def test_singleton_asyncio_concurrency() -> None:
    calls: int = 0

    async def create_resource() -> typing.AsyncIterator[str]:
        nonlocal calls
        calls += 1
        await asyncio.sleep(0)
        yield ""

    @dataclasses.dataclass(kw_only=True, slots=True)
    class SimpleCreator:
        dep1: str

    resource = providers.Resource(create_resource)
    factory_with_resource = providers.Singleton(SimpleCreator, dep1=resource.cast)

    client1, client2 = await asyncio.gather(factory_with_resource.resolve(), factory_with_resource.resolve())

    assert client1 is client2
    assert calls == 1


@pytest.mark.repeat(10)
def test_singleton_threading_concurrency() -> None:
    calls: int = 0
    lock = threading.Lock()

    def create_singleton() -> str:
        nonlocal calls
        with lock:
            calls += 1
        time.sleep(0.01)
        return ""

    singleton = providers.Singleton(create_singleton)

    with ThreadPoolExecutor(max_workers=4) as pool:
        tasks = [
            pool.submit(singleton.resolve_sync),
            pool.submit(singleton.resolve_sync),
            pool.submit(singleton.resolve_sync),
            pool.submit(singleton.resolve_sync),
        ]
        results = [x.result() for x in as_completed(tasks)]

    assert all(x == "" for x in results)
    assert calls == 1


async def test_async_singleton() -> None:
    singleton1 = await DIContainer.singleton_async()
    singleton2 = await DIContainer.singleton_async()
    singleton3 = await DIContainer.singleton_async.resolve()
    await DIContainer.singleton_async.tear_down()
    singleton4 = await DIContainer.singleton_async.resolve()

    assert singleton1 is singleton2 is singleton3
    assert singleton4 is not singleton1

    await DIContainer.tear_down()


async def test_async_singleton_override() -> None:
    singleton_async = providers.AsyncSingleton(create_async_obj, "foo")
    singleton_async.override_sync(SingletonFactory(dep1="bar"))

    result = await singleton_async.resolve()
    assert result == SingletonFactory(dep1="bar")


async def test_async_singleton_asyncio_concurrency() -> None:
    singleton_async = providers.AsyncSingleton(create_async_obj, "foo")

    results = await asyncio.gather(
        singleton_async(),
        singleton_async(),
        singleton_async(),
        singleton_async(),
        singleton_async(),
    )

    assert all(val is results[0] for val in results)


async def test_async_singleton_sync_resolve_failure() -> None:
    with pytest.raises(RuntimeError, match="AsyncSingleton cannot be resolved in an sync context."):
        DIContainer.singleton_async.resolve_sync()


async def test_singleton_async_resolve_with_async_dependencies() -> None:
    expected = await DIContainer.singleton_with_dependency.resolve()

    assert expected == await DIContainer.singleton_with_dependency.resolve()

    results = await asyncio.gather(*[DIContainer.singleton_with_dependency.resolve() for _ in range(10)])

    for val in results:
        assert val == expected

    results = await asyncio.gather(
        *[asyncio.to_thread(DIContainer.singleton_with_dependency.resolve_sync) for _ in range(10)],
    )

    for val in results:
        assert val == expected


async def test_async_singleton_teardown() -> None:
    singleton_async = providers.AsyncSingleton(create_async_obj, "foo")

    await singleton_async.resolve()
    singleton_async.tear_down_sync()

    assert singleton_async._instance is None

    await singleton_async.resolve()

    await singleton_async.tear_down()
    assert singleton_async._instance is None



import typing

from faststream import BaseMiddleware, Context, Depends
from faststream.broker.message import StreamMessage
from faststream.nats import NatsBroker, TestNatsBroker
from faststream.nats.message import NatsMessage

from that_depends import BaseContainer, container_context, fetch_context_item, providers


class ContextMiddleware(BaseMiddleware):
    async def consume_scope(
        self,
        call_next: typing.Callable[..., typing.Awaitable[typing.Any]],
        msg: StreamMessage[typing.Any],
    ) -> typing.Any:  # noqa: ANN401
        async with container_context(global_context={"request": msg}):
            return await call_next(msg)


broker = NatsBroker(middlewares=(ContextMiddleware,), validate=False)

TEST_SUBJECT = "test"


class DIContainer(BaseContainer):
    context_request = providers.Factory(
        lambda: fetch_context_item("request"),
    )


@broker.subscriber(TEST_SUBJECT)
async def index_subscriber(
    context_request: typing.Annotated[
        NatsMessage,
        Depends(DIContainer.context_request, cast=False),
    ],
    message: typing.Annotated[
        NatsMessage,
        Context(),
    ],
) -> bool:
    return message is context_request


async def test_read_main() -> None:
    async with TestNatsBroker(broker) as br:
        result = await br.request(None, TEST_SUBJECT)

        assert (await result.decode()) is True



import dataclasses
import datetime
import logging
import typing

from that_depends import BaseContainer, providers


logger = logging.getLogger(__name__)


def create_sync_resource() -> typing.Iterator[datetime.datetime]:
    logger.debug("Resource initiated")
    try:
        yield datetime.datetime.now(tz=datetime.timezone.utc)
    finally:
        logger.debug("Resource destructed")


async def create_async_resource() -> typing.AsyncIterator[datetime.datetime]:
    logger.debug("Async resource initiated")
    try:
        yield datetime.datetime.now(tz=datetime.timezone.utc)
    finally:
        logger.debug("Async resource destructed")


@dataclasses.dataclass(kw_only=True, slots=True)
class SimpleFactory:
    dep1: str
    dep2: int


async def async_factory(now: datetime.datetime) -> datetime.datetime:
    return now + datetime.timedelta(hours=1)


@dataclasses.dataclass(kw_only=True, slots=True)
class DependentFactory:
    simple_factory: SimpleFactory
    sync_resource: datetime.datetime
    async_resource: datetime.datetime


@dataclasses.dataclass(kw_only=True, slots=True)
class FreeFactory:
    dependent_factory: DependentFactory
    sync_resource: str


@dataclasses.dataclass(kw_only=True, slots=True)
class SingletonFactory:
    dep1: bool


class DIContainer(BaseContainer):
    default_scope = None
    alias = "test_container"
    sync_resource = providers.Resource(create_sync_resource)
    async_resource = providers.Resource(create_async_resource)

    simple_factory = providers.Factory(SimpleFactory, dep1="text", dep2=123)
    async_factory = providers.AsyncFactory(async_factory, async_resource.cast)
    dependent_factory = providers.Factory(
        DependentFactory,
        simple_factory=simple_factory.cast,
        sync_resource=sync_resource.cast,
        async_resource=async_resource.cast,
    )
    singleton = providers.Singleton(SingletonFactory, dep1=True)
    object = providers.Object(object())
    context_resource = providers.ContextResource(create_async_resource)



import typing

from typing_extensions import override

from that_depends.providers.base import AbstractProvider


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class Object(AbstractProvider[T_co]):
    """Provides an object "as is" without any modification.

    This provider always returns the same object that was given during
    initialization.

    Example:
        ```python
        provider = Object(1)
        result = provider.sync_resolve()
        print(result)  # 1
        ```

    """

    __slots__ = ("_obj",)

    def __init__(self, obj: T_co) -> None:
        """Initialize the provider with the given object.

        Args:
            obj (T_co): The object to be provided.

        """
        super().__init__()
        self._obj: typing.Final = obj

    @override
    async def resolve(self) -> T_co:
        return self.resolve_sync()

    @override
    def resolve_sync(self) -> T_co:
        if self._override is not None:
            return typing.cast(T_co, self._override)
        return self._obj



import datetime
import logging
import types
import typing


logger = logging.getLogger(__name__)


async def create_async_resource() -> typing.AsyncIterator[datetime.datetime]:
    logger.debug("Async resource initiated")
    try:
        yield datetime.datetime.now(tz=datetime.timezone.utc)
    finally:
        logger.debug("Async resource destructed")


def create_sync_resource() -> typing.Iterator[datetime.datetime]:
    logger.debug("Resource initiated")
    try:
        yield datetime.datetime.now(tz=datetime.timezone.utc)
    finally:
        logger.debug("Resource destructed")


class ContextManagerResource(typing.ContextManager[datetime.datetime]):
    def __enter__(self) -> datetime.datetime:
        logger.debug("Resource initiated")
        return datetime.datetime.now(tz=datetime.timezone.utc)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        logger.debug("Resource destructed")


class AsyncContextManagerResource(typing.AsyncContextManager[datetime.datetime]):
    async def __aenter__(self) -> datetime.datetime:
        logger.debug("Async resource initiated")
        return datetime.datetime.now(tz=datetime.timezone.utc)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        logger.debug("Async resource destructed")



import datetime
import typing

from faststream import Depends
from faststream.nats import NatsBroker, TestNatsBroker

from tests import container


broker = NatsBroker()

TEST_SUBJECT = "test"


@broker.subscriber(TEST_SUBJECT)
async def index_subscruber(
    dependency: typing.Annotated[
        container.DependentFactory,
        Depends(container.DIContainer.dependent_factory),
    ],
    free_dependency: typing.Annotated[
        container.FreeFactory,
        Depends(container.DIContainer.resolver(container.FreeFactory)),
    ],
    singleton: typing.Annotated[
        container.SingletonFactory,
        Depends(container.DIContainer.singleton),
    ],
    singleton_attribute: typing.Annotated[bool, Depends(container.DIContainer.singleton.dep1)],
) -> datetime.datetime:
    assert dependency.sync_resource == free_dependency.dependent_factory.sync_resource
    assert dependency.async_resource == free_dependency.dependent_factory.async_resource
    assert singleton.dep1 is True
    assert singleton_attribute is True
    return dependency.async_resource


async def test_read_main() -> None:
    async with TestNatsBroker(broker) as br:
        result = await br.request(None, TEST_SUBJECT)

        result_str = typing.cast(str, await result.decode())
        assert (
            datetime.datetime.fromisoformat(result_str.replace("Z", "+00:00"))
            == await container.DIContainer.async_resource()
        )



import asyncio
import datetime
import logging
import threading
import time
import typing
import uuid
from contextlib import AsyncExitStack, ExitStack
from unittest.mock import Mock

import pytest

from that_depends import (
    BaseContainer,
    ContextScope,
    ContextScopes,
    Provide,
    fetch_context_item,
    get_current_scope,
    inject,
    providers,
)
from that_depends.entities.resource_context import ResourceContext
from that_depends.meta import DefaultScopeNotDefinedError
from that_depends.providers import DIContextMiddleware, container_context
from that_depends.providers.context_resources import InvalidContextError, _enter_named_scope, fetch_context_item_by_type


logger = logging.getLogger(__name__)


def create_sync_context_resource() -> typing.Iterator[str]:
    logger.info("Resource initiated")
    yield f"sync {uuid.uuid4()}"
    logger.info("Resource destructed")


async def create_async_context_resource() -> typing.AsyncIterator[str]:
    logger.info("Async resource initiated")
    yield f"async {uuid.uuid4()}"
    logger.info("Async resource destructed")


class DIContainer(BaseContainer):
    alias = "context_resource_container"
    default_scope = ContextScopes.ANY
    sync_context_resource = providers.ContextResource(create_sync_context_resource)
    async_context_resource = providers.ContextResource(create_async_context_resource)
    dynamic_context_resource = providers.Selector(
        lambda: fetch_context_item("resource_type") or "sync",
        sync=sync_context_resource,
        async_=async_context_resource,
    )


class DependentDiContainer(BaseContainer):
    default_scope = ContextScopes.ANY
    dependent_sync_context_resource = providers.ContextResource(create_sync_context_resource)
    dependent_async_context_resource = providers.ContextResource(create_async_context_resource)


DIContainer.connect_containers(DependentDiContainer)


@pytest.fixture(autouse=True)
async def _clear_di_container() -> typing.AsyncIterator[None]:
    try:
        yield
    finally:
        await DIContainer.tear_down()


@pytest.fixture(params=[DIContainer.sync_context_resource, DIContainer.async_context_resource])
def context_resource(request: pytest.FixtureRequest) -> providers.ContextResource[str]:
    return typing.cast(providers.ContextResource[str], request.param)


@pytest.fixture
def sync_context_resource() -> providers.ContextResource[str]:
    return DIContainer.sync_context_resource


@pytest.fixture
def async_context_resource() -> providers.ContextResource[str]:
    return DIContainer.async_context_resource


async def test_context_resource_without_context_init(
    context_resource: providers.ContextResource[str],
) -> None:
    with pytest.raises(RuntimeError, match="Context is not set. Use container_context"):
        await context_resource.resolve()

    with pytest.raises(RuntimeError, match="Context is not set. Use container_context"):
        context_resource.resolve_sync()


async def test_context_resource(context_resource: providers.ContextResource[str]) -> None:
    async with container_context(context_resource):
        context_resource_result = await context_resource()

        assert await context_resource() is context_resource_result


def test_sync_context_resource(sync_context_resource: providers.ContextResource[str]) -> None:
    with container_context(sync_context_resource):
        context_resource_result = sync_context_resource.resolve_sync()

        assert sync_context_resource.resolve_sync() is context_resource_result


async def test_async_context_resource_in_sync_context(async_context_resource: providers.ContextResource[str]) -> None:
    with (
        pytest.raises(RuntimeError, match="Context is not set. Use container_context"),
        container_context(async_context_resource),
    ):
        await async_context_resource()


async def test_context_resource_different_context(
    context_resource: providers.ContextResource[datetime.datetime],
) -> None:
    async with container_context(context_resource):
        context_resource_instance1 = await context_resource()

    async with container_context(context_resource):
        context_resource_instance2 = await context_resource()

    assert context_resource_instance1 is not context_resource_instance2


async def test_context_resource_included_context(
    context_resource: providers.ContextResource[datetime.datetime],
) -> None:
    async with container_context(context_resource):
        context_resource_instance1 = await context_resource()
        async with container_context(context_resource):
            context_resource_instance2 = await context_resource()

        context_resource_instance3 = await context_resource()

    assert context_resource_instance1 is not context_resource_instance2
    assert context_resource_instance1 is context_resource_instance3


async def test_context_resources_overriding(context_resource: providers.ContextResource[str]) -> None:
    context_resource_mock = datetime.datetime.now(tz=datetime.timezone.utc)
    context_resource.override_sync(context_resource_mock)

    context_resource_result = await context_resource()
    context_resource_result2 = context_resource.resolve_sync()
    assert context_resource_result is context_resource_result2 is context_resource_mock

    DIContainer.reset_override_sync()
    with pytest.raises(RuntimeError, match="Context is not set. Use container_context"):
        await context_resource()


async def test_context_resources_init_and_tear_down() -> None:
    await DIContainer.init_resources()
    await DIContainer.tear_down()


def test_context_resources_wrong_providers_init() -> None:
    with pytest.raises(TypeError, match="Unsupported resource type"):
        providers.ContextResource(lambda: None)  # type: ignore[arg-type,return-value]


async def test_context_resource_with_dynamic_resource() -> None:
    async with container_context(DIContainer, global_context={"resource_type": "sync"}):
        assert (await DIContainer.dynamic_context_resource()).startswith("sync")

    async with container_context(DIContainer, global_context={"resource_type": "async_"}):
        assert (await DIContainer.dynamic_context_resource()).startswith("async")

    async with container_context(DIContainer):
        assert (await DIContainer.dynamic_context_resource()).startswith("sync")


async def test_early_exit_of_container_context() -> None:
    with pytest.raises(RuntimeError, match="No context token set for global vars, use __enter__ or __aenter__ first."):
        await container_context(global_context={"a": "a"}).__aexit__(None, None, None)
    with pytest.raises(RuntimeError, match="No context token set for global vars, use __enter__ or __aenter__ first."):
        container_context(global_context={"a": "a"}).__exit__(None, None, None)


async def test_resource_context_early_teardown() -> None:
    context: ResourceContext[str] = ResourceContext(is_async=True)
    assert context.context_stack is None
    context.tear_down_sync()
    assert context.context_stack is None


async def test_teardown_sync_container_context_with_async_resource() -> None:
    resource_context: ResourceContext[typing.Any] = ResourceContext(is_async=True)
    resource_context.context_stack = AsyncExitStack()
    message = "Cannot tear down async context in sync mode"
    with pytest.raises(RuntimeError, match=message):
        resource_context.tear_down_sync()
    with pytest.warns(RuntimeWarning, match=message):
        resource_context.tear_down_sync(raise_on_async=False)


async def test_sync_container_context_with_different_stack() -> None:
    @container_context(DIContainer)
    @inject
    def some_injected(depth: int, val: str = Provide[DIContainer.sync_context_resource]) -> str:
        if depth > 1:
            return val
        return some_injected(depth + 1)

    some_injected(1)


async def test_async_container_context_with_different_stack() -> None:
    @container_context(DIContainer)
    @inject
    async def some_injected(depth: int, val: str = Provide[DIContainer.async_context_resource]) -> str:
        if depth > 1:
            return val
        return await some_injected(depth + 1)

    await some_injected(1)


async def test_async_injection_when_resetting_resource_specific_context(
    async_context_resource: providers.ContextResource[str],
) -> None:
    """Async context resources should be able to reset the context for themselves."""

    @async_context_resource.context
    @inject
    async def _async_injected(val: str = Provide[async_context_resource]) -> str:
        assert isinstance(async_context_resource._fetch_context().context_stack, AsyncExitStack)
        return val

    async_result = await _async_injected()
    assert async_result != await _async_injected()
    assert isinstance(async_result, str)


async def test_sync_injection_when_resetting_resource_specific_context(
    sync_context_resource: providers.ContextResource[str],
) -> None:
    """Sync context resources should be able to reset the context for themselves."""

    @sync_context_resource.context
    @inject
    async def _async_injected(val: str = Provide[sync_context_resource]) -> str:
        assert isinstance(sync_context_resource._fetch_context().context_stack, ExitStack)
        return val

    @sync_context_resource.context
    @inject
    def _sync_injected(val: str = Provide[sync_context_resource]) -> str:
        assert isinstance(sync_context_resource._fetch_context().context_stack, ExitStack)
        return val

    async_result = await _async_injected()
    assert async_result != await _async_injected()
    assert isinstance(async_result, str)
    sync_result = _sync_injected()
    assert sync_result != _sync_injected()
    assert isinstance(sync_result, str)


@pytest.mark.repeat(10)
async def test_async_context_resource_asyncio_concurrency() -> None:
    calls: int = 0

    async def create_client() -> typing.AsyncIterator[str]:
        nonlocal calls
        calls += 1
        await asyncio.sleep(0)
        yield ""

    resource = providers.ContextResource(create_client)

    async def resolve_resource() -> str:
        return await resource.resolve()

    async with resource.context_async():
        await asyncio.gather(resolve_resource(), resolve_resource())

    assert calls == 1


@pytest.mark.repeat(10)
async def test_sync_context_resource_asyncio_concurrency() -> None:
    calls: int = 0

    def create_client() -> typing.Iterator[str]:
        nonlocal calls
        calls += 1
        yield ""

    resource = providers.ContextResource(create_client)

    async def resolve_resource() -> str:
        return resource.resolve_sync()

    with resource.context_sync():
        await asyncio.gather(resolve_resource(), resolve_resource())

    assert calls == 1


async def test_async_injection_when_explicitly_resetting_resource_specific_context(
    async_context_resource: providers.ContextResource[str],
) -> None:
    """Async context resources should be able to reset the context for themselves explicitly."""

    @async_context_resource.context_async()
    @inject
    async def _async_injected(val: str = Provide[async_context_resource]) -> str:
        assert isinstance(async_context_resource._fetch_context().context_stack, AsyncExitStack)
        return val

    async_result = await _async_injected()
    assert async_result != await _async_injected()
    assert isinstance(async_result, str)


async def test_sync_injection_when_explicitly_resetting_resource_specific_context(
    sync_context_resource: providers.ContextResource[str],
) -> None:
    """Sync context resources should be able to reset the context for themselves explicitly."""

    @sync_context_resource.context_async()
    @inject
    async def _async_injected(val: str = Provide[sync_context_resource]) -> str:
        assert isinstance(sync_context_resource._fetch_context().context_stack, ExitStack)
        return val

    @sync_context_resource.context_sync()
    @inject
    def _sync_injected(val: str = Provide[sync_context_resource]) -> str:
        assert isinstance(sync_context_resource._fetch_context().context_stack, ExitStack)
        return val

    async_result = await _async_injected()
    assert async_result != await _async_injected()
    assert isinstance(async_result, str)
    sync_result = _sync_injected()
    assert sync_result != _sync_injected()
    assert isinstance(sync_result, str)


async def test_async_resolution_when_explicitly_resolving(
    async_context_resource: providers.ContextResource[str],
) -> None:
    """Async context should cache resources until a new one is created."""
    async with async_context_resource.context_async():
        val_1 = await async_context_resource.resolve()
        val_2 = await async_context_resource.resolve()
        assert val_1 == val_2
        async with async_context_resource.context_async():
            val_3 = await async_context_resource.resolve()
            assert val_1 != val_3
            async with async_context_resource.context_async():
                val_4 = await async_context_resource.resolve()
                assert val_1 != val_4 != val_3
            val_5 = await async_context_resource.resolve()
            assert val_5 == val_3
        val_6 = await async_context_resource.resolve()
        assert val_6 == val_1


def test_sync_resolution_when_explicitly_resolving(
    sync_context_resource: providers.ContextResource[str],
) -> None:
    """Sync context should cache resources until a new one is created."""
    with sync_context_resource.context_sync():
        val_1 = sync_context_resource.resolve_sync()
        val_2 = sync_context_resource.resolve_sync()
        assert val_1 == val_2
        with sync_context_resource.context_sync():
            val_3 = sync_context_resource.resolve_sync()
            assert val_1 != val_3
            with sync_context_resource.context_sync():
                val_4 = sync_context_resource.resolve_sync()
                assert val_1 != val_4 != val_3
            val_5 = sync_context_resource.resolve_sync()
            assert val_5 == val_3
        val_6 = sync_context_resource.resolve_sync()
        assert val_6 == val_1


def test_sync_container_context_resolution(
    sync_context_resource: providers.ContextResource[str],
) -> None:
    """container_context should reset context for sync provider."""
    with container_context(sync_context_resource):
        val_1 = sync_context_resource.resolve_sync()
        val_2 = sync_context_resource.resolve_sync()
        assert val_1 == val_2
        with container_context(sync_context_resource):
            val_3 = sync_context_resource.resolve_sync()
            assert val_3 != val_1
        val_4 = sync_context_resource.resolve_sync()
        assert val_4 == val_1
    with pytest.raises(RuntimeError):
        sync_context_resource.resolve_sync()


async def test_async_container_context_resolution(
    async_context_resource: providers.ContextResource[str],
) -> None:
    """container_context should reset context for async provider."""
    async with container_context(async_context_resource):
        val_1 = await async_context_resource.resolve()
        val_2 = await async_context_resource.resolve()
        assert val_1 == val_2
        async with container_context(async_context_resource):
            val_3 = await async_context_resource.resolve()
            assert val_3 != val_1
        val_4 = await async_context_resource.resolve()
        assert val_4 == val_1
    with pytest.raises(RuntimeError):
        await async_context_resource.resolve()


async def test_async_global_context_resolution() -> None:
    my_global_resources = {"test_1": "test_1", "test_2": "test_2"}

    async with container_context(global_context=my_global_resources):
        for key, item in my_global_resources.items():
            assert fetch_context_item(key) == item

        async with container_context(global_context={"test_3": "test_3"}, preserve_global_context=True):
            for key, item in my_global_resources.items():
                assert fetch_context_item(key) == item

            async with container_context(global_context={"test_3": "test_3"}, preserve_global_context=False):
                for key in my_global_resources:
                    assert fetch_context_item(key) is None

            for key, item in my_global_resources.items():
                assert fetch_context_item(key) == item

        for key, item in my_global_resources.items():
            assert fetch_context_item(key) == item

    assert fetch_context_item("test_1") is None


def test_sync_global_context_resolution() -> None:
    my_global_resources = {"test_1": "test_1", "test_2": "test_2"}
    with container_context(global_context=my_global_resources):
        for key, item in my_global_resources.items():
            assert fetch_context_item(key) == item
        with container_context(global_context={"test_3": "test_3"}, preserve_global_context=True):
            for key, item in my_global_resources.items():
                assert fetch_context_item(key) == item
            with container_context(global_context={"test_3": "test_3"}, preserve_global_context=False):
                for key in my_global_resources:
                    assert fetch_context_item(key) is None
            for key, item in my_global_resources.items():
                assert fetch_context_item(key) == item
        for key, item in my_global_resources.items():
            assert fetch_context_item(key) == item

    assert fetch_context_item("test_1") is None


async def test_async_global_context_reset(async_context_resource: providers.ContextResource[str]) -> None:
    """container_context should reset async providers."""
    async with container_context(async_context_resource):
        val_1 = await async_context_resource.resolve()
        val_2 = await async_context_resource.resolve()
        assert val_1 == val_2
        async with container_context(async_context_resource):
            val_3 = await async_context_resource.resolve()
            assert val_3 != val_1
        val_4 = await async_context_resource.resolve()
        assert val_4 == val_1


def test_sync_global_context_reset(sync_context_resource: providers.ContextResource[str]) -> None:
    """container_context should reset sync providers."""
    with container_context(sync_context_resource):
        val_1 = sync_context_resource.resolve_sync()
        val_2 = sync_context_resource.resolve_sync()
        assert val_1 == val_2
        with container_context(sync_context_resource):
            val_3 = sync_context_resource.resolve_sync()
            assert val_3 != val_1
        val_4 = sync_context_resource.resolve_sync()
        assert val_4 == val_1


async def test_async_context_with_container(
    async_context_resource: providers.ContextResource[str],
    sync_context_resource: providers.ContextResource[str],
) -> None:
    """Containers should enter async context for all its providers."""
    async with DIContainer.context_async():
        val_1 = await async_context_resource.resolve()
        val_2 = await async_context_resource.resolve()
        assert val_1 == val_2
        val_1_sync = sync_context_resource.resolve_sync()
        val_2_sync = sync_context_resource.resolve_sync()
        assert val_1_sync == val_2_sync
        async with DIContainer.context_async():
            val_3 = await async_context_resource.resolve()
            val_3_sync = sync_context_resource.resolve_sync()
            assert val_3 != val_1
            assert val_3_sync != val_1_sync
        val_4 = await async_context_resource.resolve()
        val_4_sync = sync_context_resource.resolve_sync()
        assert val_4 == val_1
        assert val_4_sync == val_1_sync


def test_sync_context_with_container(
    sync_context_resource: providers.ContextResource[str],
) -> None:
    """Containers should enter sync context for all its providers."""
    with DIContainer.context_sync():
        val_1 = sync_context_resource.resolve_sync()
        val_2 = sync_context_resource.resolve_sync()
        assert val_1 == val_2
        with DIContainer.context_sync():
            val_3 = sync_context_resource.resolve_sync()
            assert val_3 != val_1
        val_4 = sync_context_resource.resolve_sync()
        assert val_4 == val_1


async def test_async_container_context_wrapper(async_context_resource: providers.ContextResource[str]) -> None:
    """Container context wrapper should correctly enter async context for wrapped function."""

    @DIContainer.context
    @inject
    async def _injected(val: str = Provide[async_context_resource]) -> str:
        return val

    assert await _injected() != await _injected()

    @DIContainer.context_async()
    @inject
    async def _explicit_injected(val: str = Provide[async_context_resource]) -> str:
        return val

    assert await _explicit_injected() != await _explicit_injected()


def test_sync_container_context_wrapper(sync_context_resource: providers.ContextResource[str]) -> None:
    """Container context wrapper should correctly enter sync context for wrapped function."""

    @DIContainer.context
    @inject
    def _injected(val: str = Provide[sync_context_resource]) -> str:
        return val

    assert _injected() != _injected()

    @DIContainer.context_sync()
    @inject
    def _explicit_injected(val: str = Provide[sync_context_resource]) -> str:
        return val

    assert _explicit_injected() != _explicit_injected()


async def test_async_context_resource_with_dependent_container() -> None:
    """Container should initialize async context resource for dependent containers."""
    async with DIContainer.context_async():
        val_1 = await DependentDiContainer.dependent_async_context_resource.resolve()
        val_2 = await DependentDiContainer.dependent_async_context_resource.resolve()
        assert val_1 == val_2


def test_sync_context_resource_with_dependent_container() -> None:
    """Container should initialize sync context resource for dependent containers."""
    with DIContainer.context_sync():
        val_1 = DependentDiContainer.dependent_sync_context_resource.resolve_sync()
        val_2 = DependentDiContainer.dependent_sync_context_resource.resolve_sync()
        assert val_1 == val_2


def test_containers_support_sync_context() -> None:
    assert DIContainer.supports_context_sync()


def test_enter_sync_context_for_async_resource_should_throw(
    async_context_resource: providers.ContextResource[str],
) -> None:
    with pytest.raises(RuntimeError):
        async_context_resource._enter_context_sync()


def test_exit_sync_context_before_enter_should_throw(sync_context_resource: providers.ContextResource[str]) -> None:
    with pytest.raises(RuntimeError):
        sync_context_resource._exit_context_sync()


async def test_exit_async_context_before_enter_should_throw(
    async_context_resource: providers.ContextResource[str],
) -> None:
    with pytest.raises(RuntimeError):
        await async_context_resource._exit_context_async()


def test_enter_sync_context_from_async_resource_should_throw(
    async_context_resource: providers.ContextResource[str],
) -> None:
    with pytest.raises(RuntimeError), ExitStack() as stack:
        stack.enter_context(async_context_resource.context_sync())


async def test_preserve_globals_and_initial_context() -> None:
    initial_context = {"test_1": "test_1", "test_2": "test_2"}

    async with container_context(global_context=initial_context):
        for key, item in initial_context.items():
            assert fetch_context_item(key) == item
        new_context = {"test_3": "test_3"}
        async with container_context(global_context=new_context, preserve_global_context=True):
            for key, item in new_context.items():
                assert fetch_context_item(key) == item
            for key, item in initial_context.items():
                assert fetch_context_item(key) == item
        for key, item in initial_context.items():
            assert fetch_context_item(key) == item
        for key in new_context:
            assert fetch_context_item(key) is None


async def test_async_context_switching_with_asyncio() -> None:
    async def slow_async_creator() -> typing.AsyncIterator[str]:
        await asyncio.sleep(0.1)
        yield str(uuid.uuid4())

    class MyContainer(BaseContainer):
        default_scope = None
        slow_provider = providers.ContextResource(slow_async_creator)

    async def _injected() -> str:
        async with MyContainer.slow_provider.context_async():
            return await MyContainer.slow_provider.resolve()

    await asyncio.gather(*[_injected() for _ in range(10)])


def test_sync_context_switching_with_threads() -> None:
    def slow_sync_creator() -> typing.Iterator[str]:
        time.sleep(0.1)
        yield str(uuid.uuid4())

    class MyContainer(BaseContainer):
        default_scope = None
        slow_provider = providers.ContextResource(slow_sync_creator)

    def _injected() -> str:
        with MyContainer.slow_provider.context_sync():
            return MyContainer.slow_provider.resolve_sync()

    threads = [threading.Thread(target=_injected) for _ in range(10)]

    for thread in threads:
        thread.start()

    for thread in threads:
        thread.join()


def test_default_named_scope_is_none() -> None:
    assert get_current_scope() is None


def test_entering_scope_sets_current_scope() -> None:
    with _enter_named_scope(ContextScopes.INJECT):
        assert get_current_scope() == ContextScopes.INJECT
    assert get_current_scope() is None


def test_entering_scope_with_container_context_sync() -> None:
    with container_context(scope=ContextScopes.INJECT):
        assert get_current_scope() == ContextScopes.INJECT
    assert get_current_scope() is None


async def test_entering_scope_with_container_context_async() -> None:
    async with container_context(scope=ContextScopes.INJECT):
        assert get_current_scope() == ContextScopes.INJECT
    assert get_current_scope() is None


def test_scoped_provider_get_scope() -> None:
    provider = providers.ContextResource(create_async_context_resource)
    assert provider.get_scope() == ContextScopes.ANY
    provider = provider.with_config(scope=ContextScopes.INJECT)
    assert provider.get_scope() == ContextScopes.INJECT


def test_scoped_container_get_scope() -> None:
    class _Container(BaseContainer): ...

    assert _Container.get_scope() is ContextScopes.ANY

    class _ScopedContainer(BaseContainer):
        default_scope = ContextScopes.INJECT

    assert _ScopedContainer.get_scope() == ContextScopes.INJECT


def test_sync_resolve_scoped_resource() -> None:
    provider = providers.ContextResource(create_sync_context_resource).with_config(scope=ContextScopes.INJECT)
    with pytest.raises(RuntimeError):
        provider.resolve_sync()

    with container_context(provider, scope=ContextScopes.INJECT):
        assert provider.resolve_sync() is not None


async def test_async_resolve_scoped_resource() -> None:
    provider = providers.ContextResource(create_async_context_resource).with_config(scope=ContextScopes.INJECT)
    with pytest.raises(RuntimeError):
        await provider.resolve()

    async with container_context(provider, scope=ContextScopes.INJECT):
        assert await provider.resolve() is not None


async def test_async_resolve_non_scoped_in_named_context() -> None:
    provider = providers.ContextResource(create_async_context_resource)
    async with container_context(provider, scope=ContextScopes.INJECT):
        assert await provider.resolve() is not None


def test_sync_resolve_non_scoped_in_named_context() -> None:
    provider = providers.ContextResource(create_sync_context_resource)
    with container_context(provider, scope=ContextScopes.INJECT):
        assert provider.resolve_sync() is not None


async def test_async_container_init_context_for_scoped_resources() -> None:
    class _Container(BaseContainer):
        async_resource = providers.ContextResource(create_async_context_resource).with_config(
            scope=ContextScopes.INJECT
        )

    async with container_context(scope=ContextScopes.INJECT):
        assert await _Container.async_resource.resolve() is not None
    with pytest.raises(RuntimeError):
        async with container_context(scope=ContextScopes.APP):
            assert await _Container.async_resource.resolve() is not None


def test_sync_container_init_context_for_scoped_resources() -> None:
    class _Container(BaseContainer):
        sync_resource = providers.ContextResource(create_sync_context_resource).with_config(scope=ContextScopes.INJECT)

    with container_context(scope=ContextScopes.INJECT):
        assert _Container.sync_resource.resolve_sync() is not None
    with pytest.raises(RuntimeError), container_context(scope=ContextScopes.APP):
        assert _Container.sync_resource.resolve_sync() is not None


async def test_sync_container_init_context_for_default_container_resources() -> None:
    class _Container(BaseContainer):
        default_scope = ContextScopes.INJECT
        sync_resource = providers.ContextResource(create_sync_context_resource)

    assert _Container.sync_resource.get_scope() == ContextScopes.INJECT
    with container_context(scope=ContextScopes.INJECT):
        assert _Container.sync_resource.resolve_sync() is not None


def test_container_with_context_resources_must_have_default_scope_set() -> None:
    with pytest.raises(DefaultScopeNotDefinedError):

        class _Container(BaseContainer):
            sync_resource = providers.ContextResource(create_sync_context_resource)


def test_providers_with_explicit_scope_ignore_default_scope() -> None:
    class _Container(BaseContainer):
        default_scope = None
        sync_resource = providers.ContextResource(create_sync_context_resource).with_config(scope=ContextScopes.INJECT)

    assert _Container.sync_resource.get_scope() == ContextScopes.INJECT


async def test_none_scoped_provider_should_not_be_resolvable_in_named_scope_async() -> None:
    provider = providers.ContextResource(create_async_context_resource).with_config(scope=None)
    async with container_context(scope=ContextScopes.INJECT):
        with pytest.raises(RuntimeError):
            await provider.resolve()


def test_none_scoped_provider_should_not_be_resolvable_in_named_scope_sync() -> None:
    provider = providers.ContextResource(create_sync_context_resource).with_config(scope=None)
    with container_context(scope=ContextScopes.INJECT), pytest.raises(RuntimeError):
        provider.resolve_sync()


def test_container_context_does_not_support_scope_any() -> None:
    with (
        pytest.raises(ValueError, match=f"{ContextScopes.ANY} cannot be entered!"),
    ):
        container_context(scope=ContextScopes.ANY)


def test_di_middleware_does_not_support_scope_any() -> None:
    with (
        pytest.raises(ValueError, match=f"{ContextScopes.ANY} cannot be entered!"),
    ):
        DIContextMiddleware(Mock(), scope=ContextScopes.ANY)


async def test_resource_context_does_not_reset_in_wrong_scope_async() -> None:
    class _Container(BaseContainer):
        default_scope = ContextScopes.REQUEST
        p_app = providers.ContextResource(create_async_context_resource).with_config(scope=ContextScopes.APP)
        p_request = providers.ContextResource(create_async_context_resource)

    async with container_context(scope=ContextScopes.APP):
        value_app_1 = await _Container.p_app.resolve()
        with pytest.raises(RuntimeError):
            await _Container.p_request.resolve()

        async with container_context(scope=ContextScopes.REQUEST):
            value_app_2 = await _Container.p_app.resolve()
            assert await _Container.p_request.resolve() is not None

            assert value_app_1 == value_app_2

        with pytest.raises(RuntimeError):
            await _Container.p_request.resolve()


def test_resource_context_does_not_set_in_wrong_scope_sync() -> None:
    class _Container(BaseContainer):
        default_scope = ContextScopes.REQUEST
        p_app = providers.ContextResource(create_sync_context_resource).with_config(scope=ContextScopes.APP)
        p_request = providers.ContextResource(create_sync_context_resource)

    with container_context(scope=ContextScopes.APP):
        value_app_1 = _Container.p_app.resolve_sync()
        with pytest.raises(RuntimeError):
            _Container.p_request.resolve_sync()

        with container_context(scope=ContextScopes.REQUEST):
            value_app_2 = _Container.p_app.resolve_sync()
            assert _Container.p_request.resolve_sync() is not None

            assert value_app_1 == value_app_2

        with pytest.raises(RuntimeError):
            _Container.p_request.resolve_sync()


async def test_strict_scope_resource_only_resolvable_in_given_scope_async() -> None:
    class _Container(BaseContainer):
        p_app = providers.ContextResource(create_async_context_resource).with_config(
            scope=ContextScopes.APP, strict_scope=True
        )
        p_request = providers.ContextResource(create_async_context_resource).with_config(
            scope=ContextScopes.REQUEST, strict_scope=True
        )

    with pytest.raises(RuntimeError):
        await _Container.p_app.resolve()

    with pytest.raises(InvalidContextError):
        await container_context(_Container.p_app, _Container.p_request).__aenter__()

    async with container_context(scope=ContextScopes.APP):
        assert await _Container.p_app.resolve() is not None
        with pytest.raises(RuntimeError):
            await _Container.p_request.resolve()

        async with container_context(scope=ContextScopes.REQUEST):
            assert await _Container.p_request.resolve() is not None
            with pytest.raises(RuntimeError):
                await _Container.p_app.resolve()


def test_strict_scope_resource_only_resolvable_in_given_scope_sync() -> None:
    class _Container(BaseContainer):
        p_app = providers.ContextResource(create_sync_context_resource).with_config(
            scope=ContextScopes.APP, strict_scope=True
        )
        p_request = providers.ContextResource(create_sync_context_resource).with_config(
            scope=ContextScopes.REQUEST, strict_scope=True
        )

    with pytest.raises(RuntimeError):
        _Container.p_app.resolve_sync()

    with pytest.raises(InvalidContextError):
        container_context(_Container.p_app, _Container.p_request).__enter__()

    with container_context(scope=ContextScopes.APP):
        assert _Container.p_app.resolve_sync() is not None
        with pytest.raises(RuntimeError):
            _Container.p_request.resolve_sync()

        with container_context(scope=ContextScopes.REQUEST):
            assert _Container.p_request.resolve_sync() is not None
            with pytest.raises(RuntimeError):
                _Container.p_app.resolve_sync()


def test_strict_scope_not_allowed_with_any_scope() -> None:
    with pytest.raises(ValueError, match=f"Cannot set strict_scope with scope {ContextScopes.ANY}."):
        providers.ContextResource(create_sync_context_resource).with_config(scope=ContextScopes.ANY, strict_scope=True)


async def test_async_resource_with_custom_scope() -> None:
    class MyScopes(ContextScopes):
        CUSTOM = ContextScope("CUSTOM")

    class _Container(BaseContainer):
        p_custom = providers.ContextResource(create_async_context_resource).with_config(scope=MyScopes.CUSTOM)

    assert _Container.p_custom.get_scope() == MyScopes.CUSTOM

    with pytest.raises(RuntimeError):
        await _Container.p_custom.resolve()

    async with container_context(_Container.p_custom, scope=MyScopes.CUSTOM):
        assert await _Container.p_custom.resolve() is not None


async def test_async_entering_container_context_correctly_handles_named_scopes() -> None:
    class _Container(BaseContainer):
        p_app = providers.ContextResource(create_async_context_resource).with_config(scope=ContextScopes.APP)
        p_request = providers.ContextResource(create_async_context_resource).with_config(scope=ContextScopes.REQUEST)

    async with container_context(_Container):
        with pytest.raises(RuntimeError):
            await _Container.p_app.resolve()

    async with container_context(_Container, scope=ContextScopes.APP):
        assert await _Container.p_app.resolve() is not None
        with pytest.raises(RuntimeError):
            await _Container.p_request.resolve()

    async with container_context(_Container, scope=ContextScopes.REQUEST):
        assert await _Container.p_request.resolve() is not None
        with pytest.raises(RuntimeError):
            await _Container.p_app.resolve()


def test_sync_entering_container_context_correctly_handles_named_scopes() -> None:
    class _Container(BaseContainer):
        p_app = providers.ContextResource(create_sync_context_resource).with_config(scope=ContextScopes.APP)
        p_request = providers.ContextResource(create_sync_context_resource).with_config(scope=ContextScopes.REQUEST)

    with container_context(_Container), pytest.raises(RuntimeError):
        _Container.p_app.resolve_sync()

    with container_context(_Container, scope=ContextScopes.APP):
        assert _Container.p_app.resolve_sync() is not None
        with pytest.raises(RuntimeError):
            _Container.p_request.resolve_sync()

    with container_context(_Container, scope=ContextScopes.REQUEST):
        assert _Container.p_request.resolve_sync() is not None
        with pytest.raises(RuntimeError):
            _Container.p_app.resolve_sync()


async def test_async_force_enter_context_for_scoped_resource() -> None:
    class _Container(BaseContainer):
        p_app = providers.ContextResource(create_async_context_resource).with_config(scope=ContextScopes.APP)

    async with _Container.p_app.context_async(force=True):
        assert await _Container.p_app.resolve() is not None

    async with _Container.context_async(force=True):
        assert await _Container.p_app.resolve() is not None


def test_sync_force_enter_context_for_scoped_resource() -> None:
    class _Container(BaseContainer):
        p_app = providers.ContextResource(create_sync_context_resource).with_config(scope=ContextScopes.APP)

    with _Container.p_app.context_sync(force=True):
        assert _Container.p_app.resolve_sync() is not None

    with _Container.context_sync(force=True):
        assert _Container.p_app.resolve_sync() is not None


async def test_async_force_enter_context_with_context_annotation() -> None:
    class _Container(BaseContainer):
        p_app = providers.ContextResource(create_async_context_resource).with_config(scope=ContextScopes.APP)

    @_Container.context(force=True)
    @inject(scope=None)
    async def _injected(val: str = Provide[_Container.p_app]) -> str:
        return val

    @_Container.p_app.context(force=True)
    @inject(scope=None)
    async def _injected_p(val: str = Provide[_Container.p_app]) -> str:
        return val

    assert await _injected() is not None
    assert await _injected_p() is not None


def test_sync_force_enter_context_with_context_annotation() -> None:
    class _Container(BaseContainer):
        p_app = providers.ContextResource(create_sync_context_resource).with_config(scope=ContextScopes.APP)

    @_Container.context(force=True)
    @inject(scope=None)
    def _injected(val: str = Provide[_Container.p_app]) -> str:
        return val

    @_Container.p_app.context(force=True)
    @inject(scope=None)
    def _injected_p(val: str = Provide[_Container.p_app]) -> str:
        return val

    assert _injected() is not None
    assert _injected_p() is not None


async def test_async_container_context_selects_context_items_on_entry() -> None:
    class _Container(BaseContainer):
        p_app = providers.ContextResource(create_async_context_resource).with_config(
            scope=ContextScopes.APP, strict_scope=True
        )
        p_request = providers.ContextResource(create_async_context_resource).with_config(
            scope=ContextScopes.REQUEST, strict_scope=True
        )

    async with container_context(scope=ContextScopes.APP):
        cc = container_context(global_context={"resource_type": "async"})

        async with container_context(scope=ContextScopes.REQUEST):
            assert get_current_scope() == ContextScopes.REQUEST
            async with cc:
                assert get_current_scope() == ContextScopes.REQUEST
                assert await _Container.p_request.resolve() is not None


def test_sync_container_context_selects_context_items_on_entry() -> None:
    class _Container(BaseContainer):
        p_app = providers.ContextResource(create_sync_context_resource).with_config(
            scope=ContextScopes.APP, strict_scope=True
        )
        p_request = providers.ContextResource(create_sync_context_resource).with_config(
            scope=ContextScopes.REQUEST, strict_scope=True
        )

    with container_context(scope=ContextScopes.APP):
        cc = container_context(global_context={"resource_type": "async"})

        with container_context(scope=ContextScopes.REQUEST):
            assert get_current_scope() == ContextScopes.REQUEST
            with cc:
                assert get_current_scope() == ContextScopes.REQUEST
                assert _Container.p_request.resolve_sync() is not None


async def test_async_container_context_wrapper_with_named_scope() -> None:
    class _Container(BaseContainer):
        p_app = providers.ContextResource(create_async_context_resource).with_config(scope=ContextScopes.APP)

    @container_context(_Container, scope=ContextScopes.APP, global_context={"resource_type": "async"})
    @inject
    async def _injected(val: str = Provide[_Container.p_app]) -> str:
        assert fetch_context_item("resource_type") == "async"
        return val

    assert await _injected() is not None


def test_sync_container_context_wrapper_with_named_scope() -> None:
    class _Container(BaseContainer):
        p_app = providers.ContextResource(create_sync_context_resource).with_config(scope=ContextScopes.APP)

    @container_context(_Container, scope=ContextScopes.APP, global_context={"resource_type": "sync"})
    @inject
    def _injected(val: str = Provide[_Container.p_app]) -> str:
        assert fetch_context_item("resource_type") == "sync"
        return val

    assert _injected() is not None


def test_container_context_must_be_called_with_arguments() -> None:
    msg = "One of context_items, scope or global_context must be provided."
    with pytest.raises(ValueError, match=msg):
        container_context()
    with pytest.raises(ValueError, match=msg):
        container_context(preserve_global_context=True)


def test_fetch_context_item_by_type() -> None:
    int_value = 3
    str_value = "value"
    with container_context(global_context={"key": str_value, "key2": int_value}):
        assert fetch_context_item_by_type(str) == str_value
        assert fetch_context_item_by_type(int) == int_value
        with pytest.raises(RuntimeError):
            fetch_context_item_by_type(float)


def test_fetch_context_item_raises() -> None:
    with pytest.raises(KeyError):
        fetch_context_item("s", raise_on_not_found=True)



import typing
from functools import partial
from typing import Annotated
from unittest.mock import Mock

from litestar import Controller, Litestar, Router, get
from litestar.di import Provide
from litestar.params import Dependency
from litestar.status_codes import HTTP_200_OK
from litestar.testing import TestClient

from that_depends import BaseContainer, providers


def bool_fn(value: bool) -> bool:
    return value


def str_fn() -> str:
    return ""


def list_fn() -> list[str]:
    return ["some"]


def int_fn() -> int:
    return 1


class SomeService:
    pass


class DIContainer(BaseContainer):
    alias = "litestar_container"
    bool_fn = providers.Factory(bool_fn, value=False)
    str_fn = providers.Factory(str_fn)
    list_fn = providers.Factory(list_fn)
    int_fn = providers.Factory(int_fn)
    some_service = providers.Factory(SomeService)


_NoValidationDependency = partial(Dependency, skip_validation=True)


class MyController(Controller):
    path = "/controller"
    dependencies = {"controller_dependency": Provide(DIContainer.list_fn)}  # noqa: RUF012

    @get(
        path="/handler",
        dependencies={
            "local_dependency": Provide(DIContainer.int_fn),
        },
    )
    async def my_route_handler(
        self,
        app_dependency: bool,
        router_dependency: str,
        controller_dependency: list[str],
        local_dependency: int,
    ) -> dict[str, typing.Any]:
        return {
            "app_dependency": app_dependency,
            "router_dependency": router_dependency,
            "controller_dependency": controller_dependency,
            "local_dependency": local_dependency,
        }

    @get(path="/mock_overriding", dependencies={"some_service": Provide(DIContainer.some_service)})
    async def mock_overriding_endpoint_handler(
        self, some_service: Annotated[SomeService, _NoValidationDependency()]
    ) -> None:
        assert isinstance(some_service, Mock)


my_router = Router(
    path="/router",
    dependencies={"router_dependency": Provide(DIContainer.str_fn)},
    route_handlers=[MyController],
)

# on the app
app = Litestar(route_handlers=[my_router], dependencies={"app_dependency": Provide(DIContainer.bool_fn)}, debug=True)


def test_litestar_endpoint_with_mock_overriding() -> None:
    some_service_mock = Mock()

    with DIContainer.some_service.override_context_sync(some_service_mock), TestClient(app=app) as client:
        response = client.get("/router/controller/mock_overriding")

    assert response.status_code == HTTP_200_OK


def test_litestar_di() -> None:
    with TestClient(app=app) as client:
        response = client.get("/router/controller/handler")
        assert response.status_code == HTTP_200_OK, response.text
        assert response.json() == {
            "app_dependency": False,
            "controller_dependency": ["some"],
            "local_dependency": 1,
            "router_dependency": "",
        }


def test_litestar_di_override_fail_on_provider_override() -> None:
    mock = 12345364758999
    with TestClient(app=app) as client, DIContainer.int_fn.override_context_sync(mock):
        response = client.get("/router/controller/handler")

    assert response.status_code == HTTP_200_OK, response.text
    assert response.json() == {
        "app_dependency": False,
        "controller_dependency": ["some"],
        "local_dependency": mock,
        "router_dependency": "",
    }


def test_litestar_di_override_fail_on_override_providers() -> None:
    mock = 12345364758999
    overrides = {
        "int_fn": mock,
    }
    with TestClient(app=app) as client, DIContainer.override_providers_sync(overrides):
        response = client.get("/router/controller/handler")

    assert response.status_code == HTTP_200_OK, response.text
    assert response.json() == {
        "app_dependency": False,
        "controller_dependency": ["some"],
        "local_dependency": mock,
        "router_dependency": "",
    }



import asyncio
import datetime
import random
import typing
import warnings
from unittest.mock import Mock

import pytest

from tests import container
from that_depends import BaseContainer, ContextScopes, Provide, inject, providers
from that_depends.injection import StringProviderDefinition


@pytest.fixture(name="fixture_one")
def create_fixture_one() -> int:
    return 1


async def _async_creator() -> typing.AsyncIterator[int]:
    yield 1


def _sync_creator() -> typing.Iterator[int]:
    yield 1


@inject
async def test_injection(
    fixture_one: int,
    simple_factory: container.SimpleFactory = Provide[container.DIContainer.simple_factory],
    dependent_factory: container.DependentFactory = Provide[container.DIContainer.dependent_factory],
    default_zero: int = 0,
) -> None:
    assert simple_factory.dep1
    assert isinstance(dependent_factory.async_resource, datetime.datetime)
    assert default_zero == 0
    assert fixture_one == 1


async def test_injection_with_overriding() -> None:
    @inject
    async def inner(
        arg1: bool,
        arg2: container.SimpleFactory = Provide[container.DIContainer.simple_factory],
    ) -> None:
        _ = arg1
        original_obj = await container.DIContainer.simple_factory()
        assert arg2.dep1 != original_obj.dep1
        assert arg2.dep2 != original_obj.dep2

    await inner(arg1=True, arg2=container.SimpleFactory(dep1="1", dep2=2))
    await inner(True, container.SimpleFactory(dep1="1", dep2=2))
    await inner(True, arg2=container.SimpleFactory(dep1="1", dep2=2))


async def test_empty_injection() -> None:
    @inject
    async def inner(_: int) -> None:
        """Do nothing."""

    warnings.filterwarnings("error")

    with pytest.raises(RuntimeWarning, match="Expected injection, but nothing found. Remove @inject decorator."):
        await inner(1)


@inject
def test_sync_injection(
    fixture_one: int,
    simple_factory: container.SimpleFactory = Provide[container.DIContainer.simple_factory],
    default_zero: int = 0,
) -> None:
    assert simple_factory.dep1
    assert default_zero == 0
    assert fixture_one == 1


def test_overriden_sync_injection() -> None:
    @inject
    def inner(
        _: container.SimpleFactory = Provide[container.DIContainer.simple_factory],
    ) -> container.SimpleFactory:
        """Do nothing."""
        return _

    factory = container.SimpleFactory(dep1="1", dep2=2)
    assert inner(_=factory) == factory


def test_sync_empty_injection() -> None:
    @inject
    def inner(_: int) -> None:
        """Do nothing."""

    warnings.filterwarnings("error")

    with pytest.raises(RuntimeWarning, match="Expected injection, but nothing found. Remove @inject decorator."):
        inner(1)


def test_type_check() -> None:
    @inject
    async def main(simple_factory: container.SimpleFactory = Provide[container.DIContainer.simple_factory]) -> None:
        assert simple_factory

    asyncio.run(main())


async def test_async_injection_with_scope() -> None:
    class _Container(BaseContainer):
        default_scope = ContextScopes.ANY
        async_resource = providers.ContextResource(_async_creator).with_config(scope=ContextScopes.INJECT)

    async def _injected(val: int = Provide[_Container.async_resource]) -> int:
        return val

    assert await inject(scope=ContextScopes.INJECT)(_injected)() == 1
    assert await inject(_injected)() == 1
    with pytest.raises(RuntimeError):
        await inject(scope=None)(_injected)()
    with pytest.raises(RuntimeError):
        await inject(scope=ContextScopes.REQUEST)(_injected)()


async def test_sync_injection_with_scope() -> None:
    class _Container(BaseContainer):
        default_scope = ContextScopes.ANY
        p_inject = providers.ContextResource(_sync_creator).with_config(scope=ContextScopes.INJECT)

    def _injected(val: int = Provide[_Container.p_inject]) -> int:
        return val

    assert inject(scope=ContextScopes.INJECT)(_injected)() == 1
    assert inject(_injected)() == 1
    with pytest.raises(RuntimeError):
        inject(scope=None)(_injected)()
    with pytest.raises(RuntimeError):
        inject(scope=ContextScopes.REQUEST)(_injected)()


def test_inject_decorator_should_not_allow_any_scope() -> None:
    with pytest.raises(ValueError, match=f"{ContextScopes.ANY} is not allowed in inject decorator."):
        inject(scope=ContextScopes.ANY)


@pytest.mark.parametrize(
    ("definition", "expected"),
    [
        ("container.provider", ("container", "provider", [])),
        ("container.provider.attr", ("container", "provider", ["attr"])),
        ("container.provider.attr1.attr2", ("container", "provider", ["attr1", "attr2"])),
        ("some.long.container.provider", ("some", "long", ["container", "provider"])),
    ],
)
def test_validate_and_extract_provider_definition_valid(definition: str, expected: tuple[str, str, list[str]]) -> None:
    """Test valid definitions and ensure the function returns the correct tuple."""
    parsed_definition = StringProviderDefinition(definition)
    result = parsed_definition._container_name, parsed_definition._provider_name, parsed_definition._attrs
    assert result == expected


@pytest.mark.parametrize(
    "definition",
    [
        "",
        "container",
        ".provider",
        "container.",
        "container..provider",
        "container.provider.",
    ],
)
def test_validate_and_extract_provider_definition_invalid(definition: str) -> None:
    """Test invalid definitions and ensure the function raises ValueError."""
    with pytest.raises(ValueError, match=f"Invalid provider definition: {definition}"):
        StringProviderDefinition(definition)


async def test_async_injection_with_string_provider_definition() -> None:
    return_value = 321321

    class _Container(BaseContainer):
        async_resource = providers.Factory(lambda: return_value)

    @inject
    async def _injected(val: int = Provide["_Container.async_resource"]) -> int:
        return val

    assert await _injected() == return_value


def test_sync_injection_with_string_provider_definition() -> None:
    return_value = 312312421

    class _Container(BaseContainer):
        sync_resource = providers.Factory(lambda: return_value)

    @inject
    def _injected(val: int = Provide["_Container.sync_resource"]) -> int:
        return val

    assert _injected() == return_value


def test_provider_string_definition_with_alias() -> None:
    return_value = 321

    class _Container(BaseContainer):
        alias = "ALIAS"
        sync_resource = providers.Factory(lambda: return_value)

    @inject
    def _injected(val: int = Provide["ALIAS.sync_resource"]) -> int:
        return val

    assert _injected() == return_value


def test_provider_string_definition_with_attr_getter() -> None:
    expected_value = 123123
    return_value = Mock()
    return_value.a = expected_value

    class _Container(BaseContainer):
        sync_resource = providers.Factory(lambda: return_value)

    @inject
    def _injected(val: int = Provide["_Container.sync_resource.a"]) -> int:
        return val

    assert _injected() == expected_value


def test_inject_with_non_existing_container() -> None:
    provider_name = "DOESNOTEXIST"

    @inject
    def _injected(val: int = Provide[f"{provider_name}.provider"]) -> None: ...

    with pytest.raises(ValueError, match=f"Container {provider_name} not found in scope!"):
        _injected()


def test_inject_with_non_existing_provider() -> None:
    container_alias = "EXIST"

    class _Container(BaseContainer):
        alias = container_alias

    provider_name = "DOESNOTEXIST"

    @inject
    def _injected(val: int = Provide[f"EXIST.{provider_name}"]) -> None: ...

    with pytest.raises(ValueError, match=f"Provider {provider_name} not found in container {container_alias}"):
        _injected()


def test_provider_resolution_with_string_definition_happens_at_runtime() -> None:
    return_value = 321

    @inject
    def _injected(val: int = Provide["_Container.sync_resource"]) -> int:
        return val

    class _Container(BaseContainer):
        sync_resource = providers.Factory(lambda: return_value)

    assert _injected() == return_value


async def test_inject_does_not_initialize_context_async() -> None:
    return_value = 3.0

    async def _async_resource() -> typing.AsyncIterator[float]:
        yield return_value

    class _Container(BaseContainer):
        provider_used = providers.ContextResource(_async_resource).with_config(scope=ContextScopes.INJECT)
        provider_unused = providers.ContextResource(_async_resource).with_config(scope=ContextScopes.INJECT)

    @inject
    async def _injected(v: float = Provide[_Container.provider_used]) -> float:
        with pytest.raises(RuntimeError):
            assert await _Container.provider_unused.resolve()

        assert isinstance(v, float)
        assert v == await _Container.provider_used.resolve()
        return v

    assert await _injected() == return_value

    with pytest.raises(RuntimeError):
        assert await _Container.provider_used.resolve()


def test_inject_does_not_initialize_context_sync() -> None:
    return_value = 3.0

    def _sync_resource() -> typing.Iterator[float]:
        yield return_value

    class _Container(BaseContainer):
        provider_used = providers.ContextResource(_sync_resource).with_config(scope=ContextScopes.INJECT)
        provider_unused = providers.ContextResource(_sync_resource).with_config(scope=ContextScopes.INJECT)

    @inject
    def _injected(v: float = Provide[_Container.provider_used]) -> float:
        with pytest.raises(RuntimeError):
            assert _Container.provider_unused.resolve_sync()
        assert isinstance(v, float)
        assert v == _Container.provider_used.resolve_sync()
        return v

    assert _injected() == return_value

    with pytest.raises(RuntimeError):
        assert _Container.provider_used.resolve_sync()


async def test_injection_initializes_context_for_parents_async() -> None:
    async def _async_resource() -> typing.AsyncIterator[float]:
        yield random.random()

    async def _async_resource_dependent(v: float) -> typing.AsyncIterator[float]:
        yield v

    class _Container(BaseContainer):
        grandparent_used = providers.ContextResource(_async_resource).with_config(scope=ContextScopes.INJECT)
        parent_used = providers.ContextResource(_async_resource_dependent, grandparent_used.cast).with_config(
            scope=ContextScopes.INJECT
        )
        provider_used = providers.ContextResource(_async_resource_dependent, parent_used.cast).with_config(
            scope=ContextScopes.INJECT
        )
        provider_unused = providers.ContextResource(_async_resource).with_config(scope=ContextScopes.INJECT)

    @inject
    async def _injected(v: float = Provide[_Container.provider_used]) -> float:
        with pytest.raises(RuntimeError):
            assert await _Container.provider_unused.resolve()
        assert isinstance(v, float)
        assert v == await _Container.provider_used.resolve()
        assert v == await _Container.parent_used.resolve()
        assert v == await _Container.grandparent_used.resolve()
        return v

    assert isinstance(await _injected(), float)
    with pytest.raises(RuntimeError):
        await _Container.provider_used.resolve()
    with pytest.raises(RuntimeError):
        await _Container.parent_used.resolve()
    with pytest.raises(RuntimeError):
        await _Container.grandparent_used.resolve()
    with pytest.raises(RuntimeError):
        await _Container.provider_unused.resolve()


def test_injection_initializes_context_for_parents_sync() -> None:
    def _sync_resource() -> typing.Iterator[float]:
        yield random.random()

    def _sync_resource_dependent(v: float) -> typing.Iterator[float]:
        yield v

    class _Container(BaseContainer):
        grandparent_used = providers.ContextResource(_sync_resource).with_config(scope=ContextScopes.INJECT)
        parent_used = providers.ContextResource(_sync_resource_dependent, grandparent_used.cast).with_config(
            scope=ContextScopes.INJECT
        )
        provider_used = providers.ContextResource(_sync_resource_dependent, parent_used.cast).with_config(
            scope=ContextScopes.INJECT
        )
        provider_unused = providers.ContextResource(_sync_resource).with_config(scope=ContextScopes.INJECT)

    @inject
    def _injected(v: float = Provide[_Container.provider_used]) -> float:
        with pytest.raises(RuntimeError):
            assert _Container.provider_unused.resolve_sync()
        assert isinstance(v, float)
        assert v == _Container.provider_used.resolve_sync()
        assert v == _Container.parent_used.resolve_sync()
        assert v == _Container.grandparent_used.resolve_sync()
        return v

    assert isinstance(_injected(), float)
    with pytest.raises(RuntimeError):
        _Container.provider_used.resolve_sync()
    with pytest.raises(RuntimeError):
        _Container.parent_used.resolve_sync()
    with pytest.raises(RuntimeError):
        _Container.grandparent_used.resolve_sync()
    with pytest.raises(RuntimeError):
        _Container.provider_unused.resolve_sync()


async def test_injection_initializes_context_for_parents_only_once_async() -> None:
    async def _async_resource() -> typing.AsyncIterator[float]:
        yield random.random()

    async def _async_resource_dependent(v: float) -> typing.AsyncIterator[float]:
        yield v

    class _Container(BaseContainer):
        grandparent = providers.ContextResource(_async_resource).with_config(scope=ContextScopes.REQUEST)
        parent = providers.ContextResource(_async_resource_dependent, grandparent.cast).with_config(
            scope=ContextScopes.REQUEST
        )
        provider = providers.ContextResource(_async_resource_dependent, parent.cast).with_config(
            scope=ContextScopes.REQUEST
        )
        provider_with_same_parent = providers.ContextResource(_async_resource_dependent, parent.cast).with_config(
            scope=ContextScopes.REQUEST
        )

    @inject(scope=ContextScopes.REQUEST)
    async def _injected(
        v_1: float = Provide[_Container.provider], v_2: float = Provide[_Container.provider_with_same_parent]
    ) -> float:
        assert isinstance(v_1, float)
        assert isinstance(v_2, float)
        assert v_1 == await _Container.provider.resolve()
        assert v_1 == await _Container.parent.resolve()
        assert v_1 == await _Container.grandparent.resolve()
        assert v_1 == v_2
        return v_1 + v_2

    assert isinstance(await _injected(), float)


def test_injection_initializes_context_for_parents_only_once_sync() -> None:
    def _sync_resource() -> typing.Iterator[float]:
        yield random.random()

    def _sync_resource_dependent(v: float) -> typing.Iterator[float]:
        yield v

    class _Container(BaseContainer):
        default_scope = ContextScopes.APP
        grandparent = providers.ContextResource(_sync_resource).with_config(scope=ContextScopes.ANY)
        parent = providers.ContextResource(_sync_resource_dependent, grandparent.cast).with_config(
            scope=ContextScopes.APP
        )
        provider = providers.ContextResource(_sync_resource_dependent, parent.cast).with_config(scope=ContextScopes.APP)
        provider_with_same_parent = providers.ContextResource(_sync_resource_dependent, parent.cast).with_config(
            scope=ContextScopes.APP
        )

    @inject(scope=ContextScopes.APP)
    def _injected(
        v_1: float = Provide[_Container.provider], v_2: float = Provide[_Container.provider_with_same_parent]
    ) -> float:
        assert isinstance(v_1, float)
        assert isinstance(v_2, float)
        assert v_1 == _Container.provider.resolve_sync()
        assert v_1 == _Container.parent.resolve_sync()
        assert v_1 == _Container.grandparent.resolve_sync()
        assert v_1 == v_2
        return v_1 + v_2

    assert isinstance(_injected(), float)


async def test_inject_scope_does_not_init_context_for_out_of_scope_dependents_async() -> None:
    async def _async_resource() -> typing.AsyncIterator[float]:
        yield random.random()

    class _Container(BaseContainer):
        parent = providers.ContextResource(_async_resource).with_config(scope=ContextScopes.INJECT)
        child = providers.Factory(lambda x: x, parent.cast)

    @inject(scope=ContextScopes.APP)
    async def _injected(v: float = Provide[_Container.child]) -> float:
        return v

    with pytest.raises(RuntimeError):
        await _injected()

    async with _Container.parent.context_async(force=True):
        parent_value = await _Container.parent.resolve()
        assert await _injected() == parent_value


async def test_inject_scope_does_not_init_context_for_out_of_scope_dependents_sync() -> None:
    def _sync_resource() -> typing.Iterator[float]:
        yield random.random()

    class _Container(BaseContainer):
        parent = providers.ContextResource(_sync_resource).with_config(scope=ContextScopes.INJECT)
        child = providers.Factory(lambda x: x, parent.cast)

    @inject(scope=ContextScopes.APP)
    def _injected(v: float = Provide[_Container.child]) -> float:
        return v

    with pytest.raises(RuntimeError):
        _injected()

    with _Container.parent.context_sync(force=True):
        parent_value = _Container.parent.resolve_sync()
        assert _injected() == parent_value
